"""
Controls the TribBas Model. Highly eperimental and evolves really rapidly So no doc so far sns.
No documentation so far, B.G. is still testing it.
"""
from lsdtt_xtensor_python import run_TribBas_to_steady_state, prebuild_TribBas, run_TribBas # Old stuff
from lsdtt_xtensor_python import LSDTribBas_cpp
import numba
import numpy as np
import pandas as pd
from lsdtopytools import raster_loader as rl, geoconvtools as gc
from lsdtopytools.lsdtopytools_utilities import save_to_database, load_from_database, load_metadata_from_database
from . import LSDDEM


class LSDTribBas(object):
	"""
		The LSDTribBas class provides a high level interface to the LSDTribBas Model in LSDTopoTools.
		May disappear at some points but I need that for automation.
		Will develop if needed.
		Boris
	"""

	def __init__(self, raster_name = "buzau.bil", source_csv = "buzau_Wsources.csv", output_name = "Clegg.h5", outlet_x=0, outlet_y =0, search_node_radius = 100, target_node = 20, MC_iterations = 50, skip = 1, sigma =2, m = 1.4, n = 3.11, A0 = 1, verbose = True, already_preprocessed = True):
		"""
		Constructor

		"""
		# First: Do you want the model to talk much?
		self.verbose = verbose

		#Loading raster
		print("Loading raster into the system") if self.verbose else 0
		A = rl.load_raster(raster_name)
		print("Raster loaded !") if self.verbose else 0
		if(~already_preprocessed):
			B = LSDDEM(file_name = raster_name)
			B.PreProcessing()
			A["array"] = B.cppdem.get_PP_raster()

		print("Loading my sources") if self.verbose else 0
		# I will need the sources
		sources = pd.read_csv(source_csv)
		# Trimming my sources as they are extracted from a larger raster
		SOX = sources['x'][(sources['x'] > A['x_min']) & (sources['x']<A['x_max'])].values
		SOY = sources['y'][(sources['y'] > A['y_min']) & (sources['y']<A['y_max'])].values
		print("Sources ready to be ingested") if self.verbose else 0


		# Prebuilding the model
		self.n = n
		self.m = m
		self.A0 = A0

		# HD5 file
		self.hd5_name = output_name
		self.run_default_ID = 0

		print("Now prebuilding your model: \n-> getting all the flow informations\n-> ingesting the sources to extract a river network\n-> Calculate initial ksn using Mudd et al., 2014 (JGR)\n-> Order and format my nodes\nIt can take a bit of time...") if self.verbose else 0
		self.TBcpp = LSDTribBas_cpp(A["nrows"], A["ncols"], A["x_min"], A["y_min"], A["res"], A["nodata"][0], A["array"],SOX, SOY, m, n, A0, outlet_x, outlet_y, search_node_radius, target_node, MC_iterations, skip, 10, sigma)
		print("Model constructed!") if self.verbose else 0
		self.n_node = self.TBcpp.m_chi().shape[0]

		self.base_param = [A["nrows"], A["ncols"], A["x_min"], A["y_min"], A["res"], A["nodata"][0], A["array"],SOX, SOY, m, n, A0, outlet_x, outlet_y, search_node_radius, target_node, MC_iterations, skip, 10, sigma]

		print("Saving the initial state to hd5...")
		storer = pd.HDFStore(self.hd5_name)
		storer.put("initial",pd.DataFrame({"flow_distance": self.TBcpp.flow_distance(), "elevation": self.TBcpp.elevation()}),format = "t")
		storer.close()

		# self.K_to_SS = {}

	def run_model(self,K_field = 5e-8, uplift_field = 0.001, tolerance_p = 0.95, tolerance_delta = 0.5, timestep = 1000, final_tol = 0.95, save_dir = "./", n_timestep = "auto", save_step = 10, min_n_timestep = 25 
		, max_n_timestep = 10000, run_save_ID = None):
		"""
			Just run the model
			K and U fields can be scalar or ndarray
		"""

		# getting the uplift and K to the right format
		if(~isinstance(uplift_field,np.ndarray)):
			uplift_field = np.zeros(self.n_node) + uplift_field
		if(~isinstance(K_field,np.ndarray)):
			K_field = np.zeros(self.n_node) + K_field

		if(isinstance(n_timestep,str)):
			if(n_timestep == "auto"):
				results = self.TBcpp.run_model(timestep, save_step,tolerance_p, tolerance_delta, min_n_timestep , uplift_field, K_field, max_n_timestep)
		else:
			results = self.TBcpp.run_model(timestep, save_step,tolerance_p, tolerance_delta, n_timestep, uplift_field, K_field, n_timestep)

		if(run_save_ID is None):
			outID = "run_" + str(self.run_default_ID + 1)
			print("Defaulting the name to %s" %(outID))
		else:
			outID = str(run_save_ID)

		storer = pd.HDFStore(self.hd5_name)
		storer.put(outID,pd.DataFrame(results), format="t")
		storer.close()


	def m_and_ns(self, min_n = 1, min_m = 1, step = 0.1, nsteps = 40, min_K = 1e-9, range_of_K = [1e-9,1e-8,1e-7,1e-6,1e-5,1e-4,1e-3], uplift_field = 0.001, tolerance_p = 0.95, tolerance_delta = 0.5, timestep = 1000, 
		final_tol = 0.95, save_dir = "./", min_n_timestep = 25 , max_n_timestep = 10000, run_save_ID = None, interval_of_test = 1000):
		"""
		try stuff on best fit m and n
		"""	
		from lsdtt_xtensor_python import get_median_profile

		print("Initialising the model, it will take a while as I am calculating m_chi for all m/n combination")
		# initialising the ranges
		range_m = []
		range_n = []
		m_chi_in_theta = {} # Dictionnary to avoid duplicity of data
		chi_in_theta = {}
		compmet = {}

		this_U = uplift_field
		unique_ID_m_and_ns = 0
		# Output managers
		analysis_names = []# 

		# Getting the ranges of m n and m/n
		for i in range(nsteps):
			range_m.append(min_m + step * i)
			range_n.append(min_n + step * i)
		

		for tn in range_n:
			for tm in range_m:
				if(tm<tn):
					ratio = tm/tn
					ratio = str(ratio)[:4]
					if(~(ratio in m_chi_in_theta)):
						# print("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",ratio, tm, tn)
						temp_OB = LSDTribBas_cpp(self.base_param[0], self.base_param[1], self.base_param[2], self.base_param[3], self.base_param[4], self.base_param[5],self.base_param[6],self.base_param[7], self.base_param[8], range_m[-1] , range_n[-1], self.base_param[11], self.base_param[12], self.base_param[13], self.base_param[14], self.base_param[15], self.base_param[16], self.base_param[17], self.base_param[18], self.base_param[19])
						m_chi_in_theta[ratio] = temp_OB.m_chi()
						chi_in_theta[ratio] = temp_OB.chi()

		# for key,val in m_chi_in_theta.items():
		# 	print(key)

		print("Model initialised!")

		save_to_database(self.hd5_name,"reference_mchi", pd.DataFrame(m_chi_in_theta))


		print("Ok Let's roll")
		this_K = min_K
		for this_K in (range_of_K):
			# getting the uplift and K to the right format
			if(~isinstance(uplift_field,np.ndarray)):
				uplift_field = np.zeros(self.n_node) + uplift_field

			K_field = np.zeros(self.n_node) + this_K

			for m in range_m:
				for n in range_n:
					if(m<n): # Checking that m/n < 1
						print("Calculating m: %s and n: %s" %(m,n))
						self.TBcpp.set_m_and_n(m,n)
						# Running the model to steady state
						results = pd.DataFrame(self.TBcpp.run_model(timestep, 100000,tolerance_p, tolerance_delta, min_n_timestep , uplift_field, K_field, max_n_timestep))
						maxi = 0
						for col in results.columns :
							if( float(col) > maxi):
								maxi = col
						# Getting the results
						this_elev_array = results[maxi].values
						temp_df = pd.DataFrame({"flow_distance": self.TBcpp.flow_distance(), "initial_elevation": self.TBcpp.elevation(), "SS_elevation": this_elev_array})
						temp_df = temp_df.sort_values("flow_distance")

						# Getting the median values
						compa = get_median_profile(temp_df["flow_distance"].values, temp_df["SS_elevation"].values - temp_df["initial_elevation"].values,interval_of_test, 1)
						# saving the values
						kalak = str(m) + "_" + str(n)
						compmet[kalak + "_elev"] = compa["Y"]

						# dealing with median m_chi
						this_ratio = str(m/n)[:4]
						this_mchi = self.TBcpp.first_order_m_chi_from_custarray(chi_in_theta[this_ratio],this_elev_array)
						compa_2 = get_median_profile(temp_df["flow_distance"].values, m_chi_in_theta[this_ratio] - this_mchi, interval_of_test, 1)
						compmet[kalak + "_mchi"] = compa_2["Y"]



			compmet["X"] = compa["X"] #  for plotting. Because flow distance is always the same, X is as well
			df = pd.DataFrame(compmet)
			# temp for testing

			# df.to_csv("test_of_mn.csv", index = False)
			analysis_names.append("mn_test_" + str(unique_ID_m_and_ns))
			save_to_database(self.hd5_name,"mn_test_" + str(unique_ID_m_and_ns), df, metadata = {"uplift_field": this_U, "K_field": this_K})
			# storer = pd.HDFStore(self.hd5_name)
			# storer.put("U%s_K%s"%(this_U,this_K),df, format="t")
			# storer.close()
			unique_ID_m_and_ns += 1
			# this_K = 10 ** (np.log10(this_K) + 0.1)
		# Saving lists of m,n,k and U that have been effectively tested
		save_to_database(self.hd5_name,"mn_test_metadata", pd.DataFrame({"test": [0,0,0,1,2]}), metadata = {"analysis_names": analysis_names})



	def TODO(self):
		print("""
			TODO:
			- Adapt the run_model's output to get the stat of fit for each timestep
			- implement m_chi calculation and comparison
			- Simplification/multithreading
			- Dorce K or U to behave by segments
			""")










################## OLD ROUTINES I USED FOR CHRISTMAS. BASIC AND BRUTFORCE

	# def hunt_my_K(self,initial_k = 5e-5, reference_k_step = 1e-5, uplift_field = 0.001, min_iteration_per_run = 20, tolerance_p = 0.95, tolerance_delta = 0.5, timestep = 1000, final_tol = 0.95,  final_tol_delta = 0.5, save_dir = "./", max_iterations_per_test= 500, max_global_iterations = 100 ):
	# 	"""
	# 		Iterative method to force K refining
	# 	"""
	# 	# We force an uplift field to remain constant
	# 	U = np.zeros(self.prebuilt["elevation"].shape[0]) + uplift_field
	# 	# Initition of erodibility
	# 	K = np.zeros(self.prebuilt["elevation"].shape[0]) + initial_k
	# 	# Initialization of adaptative erodibility step
	# 	refK = np.zeros(self.prebuilt["elevation"].shape[0]) + reference_k_step
	# 	# sign of K last time
	# 	signk = np.zeros(self.prebuilt["elevation"].shape[0])
	# 	last_sign_k = np.zeros(self.prebuilt["elevation"].shape[0])
	# 	how_many_time_unchanged = np.zeros(self.prebuilt["elevation"].shape[0])


	# 	run = {}

	# 	catched = False

	# 	cpt = 0
	# 	this_prebuilt = self.prebuilt.copy()

	# 	bdf = pd.DataFrame(data = self.prebuilt)
	# 	bdf.to_csv(save_dir+"base.csv", index = False)



	# 	print("Chasing my K!")
	# 	while(catched == False and cpt<max_global_iterations): # safety to 100
	# 		cpt+=1 # ᶘ °㉨°ᶅ
	# 		print("MODEL RUN %s"%(cpt), end="\r")

	# 		# Run little K
	# 		results = run_TribBas_to_steady_state(self.m,self.n, timestep, 10000, tolerance_p, tolerance_delta, 20, U, K, this_prebuilt, max_iterations_per_test)

	# 		# print("FOUND STEADY-STATE, REFINING MY K")

	# 		# get the latest result
	# 		testk = 0
	# 		maxk = 0
	# 		for key,val in results.items():
	# 			testk = int(key) # string to int
	# 			print(key)
	# 			if(testk>maxk):
	# 				maxk = testk
	# 		maxk = str(int(maxk)) # getting the right format
	# 		maxk = maxk.split(".")[0] # Making sure no decimal are there
	# 		print("DEBUG::maxk:", maxk)
	# 		new_elev = results[maxk]
	# 		base_elev = np.copy(self.prebuilt["elevation"])
	# 		print("DEBUG::mean_Baseelev:", np.mean(base_elev))

	# 		n_tol = new_elev[np.abs(new_elev-base_elev)<final_tol_delta].shape[0]
	# 		n_tot = new_elev.shape[0]
	# 		if(n_tol/n_tot>final_tol):
	# 			catched = True
	# 			print("I THINK I FOUND MY K !!!")
	# 		else:
	# 			print("REFINING MY K ...")
	# 			signk[new_elev<base_elev] = -1
	# 			signk[new_elev>base_elev] = 1
	# 			signk[np.abs(new_elev-base_elev)<final_tol_delta] = 0

	# 			if(cpt==1): # first run is 1 (incrementation at the beginning of the loop, see ᶘ °㉨°ᶅ)
	# 				last_sign_k = np.copy(signk)

	# 			K, refK, how_many_time_unchanged = self.adjust_my_K_wild(base_elev,new_elev,refK,signk,last_sign_k,final_tol,K,how_many_time_unchanged)

	# 			print("I ADAPTED %s/%s K."%(signk[signk!=0].shape[0],signk.shape[0]))

	# 		df = pd.DataFrame(data = {"K":K, "elevation": new_elev})

	# 		name = str(cpt)
	# 		while(len(name)<4):
	# 			name = "0" +name

	# 		df.to_csv(save_dir + "attempt_%s_K.csv"%(name))
	# 		this_prebuilt["elevation"] = new_elev
	# 		last_sign_k = np.copy(signk)
	# 	return K



	# # Temp function, will move to cpp
	# @numba.jit()
	# def adjust_my_K_wild(self, base_elev,new_elev,refK,signk,last_sign_k,final_tol,K,how_many_time_unchanged):

	# 	for i in range(base_elev.shape[0]):
	# 		# Need adjustment?
	# 		if(abs(new_elev[i]-base_elev[i])>final_tol):
	# 			if(signk[i] != last_sign_k[i]):
	# 				refK[i] = refK[i]/2
	# 				how_many_time_unchanged[i] = 0
	# 			else:
	# 				how_many_time_unchanged[i] = how_many_time_unchanged[i]+1
	# 				if(how_many_time_unchanged[i]>10):
	# 					refK[i] = refK[i]*1.5
	# 					how_many_time_unchanged[i] = 0



	# 			last_K = K[i]
	# 			K[i] += signk[i]*refK[i]
	# 			if(K[i]<=0):
	# 				K[i] = refK[i]/2 # HACKY WAY HERE
	# 				refK[i] = refK[i]/2 # HACKY WAY HERE


	# 	return K, refK,how_many_time_unchanged

	# def hunt_my_U(self,initial_uplift = 0.001, reference_uplift_step = 0.0001, min_iteration_per_run = 20 , K_field = 0.001, tolerance_p = 0.95, tolerance_delta = 0.5, timestep = 1000, final_tol = 0.95,  final_tol_delta = 0.5, save_dir = "./", max_iterations_per_test= 500, max_global_iterations = 100 ):
	# 	"""
	# 		Iterative method to force K refining
	# 	"""
	# 	# We force an uplift field to remain constant
	# 	U = np.zeros(self.prebuilt["elevation"].shape[0]) + initial_uplift
	# 	# Initition of erodibility
	# 	K = np.zeros(self.prebuilt["elevation"].shape[0]) + K_field
	# 	# Initialization of adaptative erodibility step
	# 	refU = np.zeros(self.prebuilt["elevation"].shape[0]) + reference_uplift_step
	# 	# sign of K last time
	# 	signU = np.zeros(self.prebuilt["elevation"].shape[0])
	# 	last_sign_k = np.zeros(self.prebuilt["elevation"].shape[0])
	# 	how_many_time_unchanged = np.zeros(self.prebuilt["elevation"].shape[0])


	# 	run = {}

	# 	catched = False

	# 	cpt = 0
	# 	this_prebuilt = self.prebuilt.copy()

	# 	bdf = pd.DataFrame(data = self.prebuilt)
	# 	bdf.to_csv(save_dir+"base.csv", index = False)



	# 	print("Chasing my theoretical Uplift field!")
	# 	while(catched == False and cpt<max_global_iterations): # safety to 100
	# 		cpt+=1 # ᶘ °㉨°ᶅ
	# 		print("MODEL RUN %s"%(cpt))

	# 		# Run little K
	# 		results = run_TribBas_to_steady_state(self.m,self.n, timestep, 10000, tolerance_p, tolerance_delta, min_iteration_per_run, U, K, this_prebuilt, max_iterations_per_test)

	# 		print("FOUND STEADY-STATE, REFINING MY K")

	# 		# get the latest result
	# 		testU = 0
	# 		maxU = 0
	# 		for key,val in results.items():
	# 			testU = int(key) # string to int
	# 			print(key)
	# 			if(testU>maxU):
	# 				maxU = testU
	# 		maxU = str(int(maxU)) # getting the right format
	# 		maxU = maxU.split(".")[0] # Making sure no decimal are there
	# 		# print("DEBUG::maxU:", maxU)
	# 		new_elev = results[maxU]
	# 		base_elev = np.copy(self.prebuilt["elevation"])
	# 		# print("DEBUG::mean_Baseelev:", np.mean(base_elev))

	# 		n_tol = new_elev[np.abs(new_elev-base_elev)<final_tol_delta].shape[0]
	# 		n_tot = new_elev.shape[0]
	# 		if(n_tol/n_tot>final_tol):
	# 			catched = True
	# 			print("I THINK I FOUND MY U :)")
	# 		else:
	# 			print("REFINING MY U :)")
	# 			signU[new_elev<base_elev] = 1
	# 			signU[new_elev>base_elev] = -1
	# 			signU[np.abs(new_elev-base_elev)<final_tol_delta] = 0

	# 			if(cpt==1): # first run is 1 (incrementation at the beginning of the loop, see ᶘ °㉨°ᶅ)
	# 				last_sign_U = np.copy(signU)

	# 			U, refU, how_many_time_unchanged = self.adjust_my_U_wild(base_elev,new_elev,refU,signU,last_sign_U,final_tol,U,how_many_time_unchanged)

	# 			print("I ADAPTED %s/%s U."%(signU[signU!=0].shape[0],signU.shape[0]))

	# 		df = pd.DataFrame(data = {"U":U, "elevation": new_elev})

	# 		name = str(cpt)
	# 		while(len(name)<4):
	# 			name = "0" +name

	# 		df.to_csv(save_dir + "attempt_%s_U.csv"%(name))
	# 		this_prebuilt["elevation"] = new_elev
	# 		last_sign_U = np.copy(signU)

	# 	return U



	# # Temp function, will move to cpp
	# @numba.jit()
	# def adjust_my_U_wild(self, base_elev,new_elev,refU,signU,last_sign_U,final_tol,U,how_many_time_unchanged):

	# 	for i in range(base_elev.shape[0]):
	# 		# Need adjustment?
	# 		if(abs(new_elev[i]-base_elev[i])>final_tol):
	# 			if(signU[i] != last_sign_U[i]):
	# 				refU[i] = refU[i]/2
	# 				how_many_time_unchanged[i] = 0
	# 			else:
	# 				how_many_time_unchanged[i] = how_many_time_unchanged[i]+1
	# 				if(how_many_time_unchanged[i]>10):
	# 					refU[i] = refU[i]*1.5
	# 					how_many_time_unchanged[i] = 0



	# 			last_U = U[i]
	# 			U[i] += signU[i]*refU[i]
	# 			if(U[i]<=0):
	# 				U[i] = refU[i]/2 # HACUY WAY HERE
	# 				refU[i] = refU[i]/2 # HACUY WAY HERE


	# 	return U, refU,how_many_time_unchanged















#!/usr/bin/env python
"""
Command-line tool to control the concavity constraining tools
Mudd et al., 2018
So far mostly testing purposes
B.G.
"""
from lsdtopytools import LSDDEM # I am telling python I will need this module to run.
from lsdtopytools import argparser_debug as AGPD # I am telling python I will need this module to run.
from lsdtopytools import quickplot as qp, quickplot_movern as qmn # We will need the plotting routines
import time as clock # Basic benchmarking
import sys # manage the argv

def main_concavity():
	# Here are the different parameters and their default value fr this script
	default_param = AGPD.get_common_default_param()
	default_param["quick_movern"] = False
	default_param["X"] = None
	default_param["Y"] = None

	default_param = AGPD.ingest_param(default_param, sys.argv)


	# Checking the param
	if(isinstance(default_param["X"],str)):
		X = [float(default_param["X"])]
		Y = [float(default_param["Y"])]
	else:
		X = [float(i) for i in default_param["X"]]
		Y = [float(i) for i in default_param["Y"]]


	if(default_param["help"] or len(sys.argv)==1):
		print("""
			This command-line tool run concavity analysis tools from LSDTopoTools.
			Description of the algorithms in Mudd et al., 2018 -> https://www.earth-surf-dynam.net/6/505/2018/
			To use, run the script with relevant options, for example:
				lsdtt-concavity-tools.py file=myraster.tif quick_movern X=5432 Y=78546

			option available:
				file: name of the raster (file=name.tif)
				path: path to the file (default = current folder)
				quick_movern: run disorder metrics and plot a result figure (you jsut need to write it)
				X: X coordinate of the outlet (So far only single basin is supported as this is an alpha tool)
				Y: Y coordinate of the outlet (So far only single basin is supported as this is an alpha tool)
				help: if written, diplay this message. Documentation soon to be written.
			""")
		quit()


	print("Welcome to the command-line tool to constrain your river network concavity. Refer to Mudd et al., 2018 -> https://www.earth-surf-dynam.net/6/505/2018/ for details about these algorithms.")
	print("Let me first load the raster ...")
	mydem = LSDDEM(file_name = default_param["file"], path=default_param["path"], already_preprocessed = False, verbose = False)
	print("Got it. Now dealing with the depressions ...")
	mydem.PreProcessing(filling = True, carving = True, minimum_slope_for_filling = 0.0001) # Unecessary if already preprocessed of course.
	print("Done! Extracting the river network")
	mydem.ExtractRiverNetwork( method = "area_threshold", area_threshold_min = 1000)



	print("I have some rivers for you! Defining the watershed of interest...")
	mydem.DefineCatchment( method="from_XY", X_coords = X, Y_coords = Y, test_edges = False, coord_search_radius_nodes = 25, coord_threshold_stream_order = 1)
	print("I should have it now.")
	print("I got all the common info, now I am running what you want!")

	if(default_param["quick_movern"]):
		print("Initialising Chi-culations (lol)")
		mydem.GenerateChi()
		print("Alright, getting the disorder metrics for each chi values. LSDTopoTools can split a lot of messages, sns.")
		mydem.cppdem.calculate_movern_disorder(0.15, 0.05, 17, 1, 1000) # start theta, delta, n, A0, threashold
		print("I am done, plotting the results now")
		qmn.plot_disorder_results(mydem, legend = False, normalise = True, cumulative_best_fit = False)
		qp.plot_check_catchments(mydem)
		qmn.plot_disorder_map(mydem, cmap = "RdBu_r")
		print("FInished with quick disorder metric")

	print("Finished!")




#!/usr/bin/env python
"""
Command-line tool to control the concavity constraining tools
Mudd et al., 2018
So far mostly testing purposes
B.G.
"""
from lsdtopytools import LSDDEM, raster_loader as rl # I am telling python I will need this module to run.
from lsdtopytools import argparser_debug as AGPD # I am telling python I will need this module to run.
from lsdtopytools import quickplot as qp, quickplot_movern as qmn # We will need the plotting routines
import time as clock # Basic benchmarking
import sys # manage the argv

def PreProcess():
	# Here are the different parameters and their default value fr this script
	default_param = AGPD.get_common_default_param()
	default_param["breach"] = False
	default_param["fill"] = False
	default_param["min_slope"] = 0.0001



	default_param = AGPD.ingest_param(default_param, sys.argv)

	if(default_param["help"] or len(sys.argv)==1):
		print("""
			This command-line tool provides basic Preprocessing functions for a raster.
			Breaching is achieved running https://doi.org/10.1002/hyp.10648 algorithm (Lindsay2016), with an implementation from RICHDEM
			Filling is from Wang and liu 2006. I think MDH implemented the code.
			It also "clean" the raster and tidy the nodata in it.
			To use, run the script with relevant options, for example:
			-> filling and breaching
				lsdtt-depressions file=myraster.tif fill breach 0.00001 

			option available:
				file: name of the raster (file=name.tif)
				path: path to the file (default = current folder)
				breach: want to breach?
				fill: want to fill?
				min_slope: imply a min slope for filling

			""")
		quit()


	print("Welcome to the command-line tool to .")
	print("Let me first load the raster ...")
	mydem = LSDDEM(file_name = default_param["file"], path=default_param["path"], already_preprocessed = False, verbose = False)
	print("Got it. Now dealing with the depressions ...")
	mydem.PreProcessing(filling = default_param["fill"], carving = default_param["breach"], minimum_slope_for_filling = default_param["min_slope"]) # Unecessary if already preprocessed of course.

	print("Saving the raster! same name basically, but with _PP at the end")
	rl.save_raster(mydem.cppdem.get_PP_raster(),mydem.x_min,mydem.x_max,mydem.y_max,mydem.y_min,mydem.resolution,mydem.crs,mydem.path+mydem.prefix + "_PP.tif", fmt = 'GTIFF')

def Polyfit_Metrics():

	# Here are the different parameters and their default value fr this script
	default_param = AGPD.get_common_default_param()
	default_param["breach"] = False
	default_param["fill"] = False
	default_param["window_radius"] = 30

	default_param["average_elevation"] = False
	default_param["slope"] = False
	default_param["aspect"] = False
	default_param["curvature"] = False
	default_param["planform_curvature"] = False
	default_param["profile_curvature"] = False
	default_param["tangential_curvature"] = False
	default_param["TSP"] = False



	default_param = AGPD.ingest_param(default_param, sys.argv)

	if(default_param["help"] or len(sys.argv)==1):
		print("""
			This command-line tool provides basic Preprocessing functions for a raster.
			Breaching is achieved running https://doi.org/10.1002/hyp.10648 algorithm (Lindsay2016), with an implementation from RICHDEM
			Filling is from Wang and liu 2006. I think MDH implemented the code.
			It also "clean" the raster and tidy the nodata in it.
			To use, run the script with relevant options, for example:
			lsdtt-polyfits file=myraster.tif slope window_radius 17
			option available:
				file: name of the raster (file=name.tif)
				path: path to the file (default = current folder)
				TODO
			""")
		quit()


	print("Welcome to the command-line tool to .")
	print("Let me first load the raster ...")
	mydem = LSDDEM(file_name = default_param["file"], path=default_param["path"], already_preprocessed = True, verbose = False)

	mydem.get_polyfit_metrics(window_radius = default_param["window_radius"], average_elevation = default_param["average_elevation"], slope = default_param["slope"], aspect = default_param["aspect"], curvature = default_param["curvature"], planform_curvature = default_param["planform_curvature"], profile_curvature = default_param["profile_curvature"], tangential_curvature = default_param["tangential_curvature"], TSP = default_param["TSP"], save_to_rast = True)









"""
This files will manage all the shape-file related operation. Espescially I/O probably.

Authors: B.G
"""
import fiona, shapely as shp, geopandas as gpd
import numpy as np
import pandas as pd

class lsdSHP(object):
	"""
		lsdSHP class contains tools to interact with shapefiles and make them ingestable to 
	"""
	def __init__(self, file_name, path = "./", verbose = True):
		"""
			Constructor of the lsdSHP object.
			Arguments:
				file_name (str): the name of the file including the main extention (e.g. "gabulan.shp", "falafel.json")
				path (str): path to the file to read/write (default=current folder)
				verbose (bool): The code is quite chatty, you may want to ask it to keep it quiet.
			Returns:
				a brand new lsdSHP object Yeah
			Authors:
				B.G.
			Date:
				04/03/2019
		"""
		# Ignore that line
		super(lsdSHP, self).__init__()
		# File to load
		self.file_name = file_name
		# and its path
		self.path = path

		# Loading the file with geopandas
		self.shpdf = gpd.read_file(path+self.file_name)

		# Loading a bunch of useful info to have
		self.data_info = {"type": self.shpdf.geometry.iloc[0], "n_elements": self.shpdf}

		# Verbosity of the code
		self.verbose = verbose

		# Getting the prefix for saving all your files:
		self.prefix = str("".join(self.file_name.split(".")[:-1])) # -> prefix = base name without extention




		
	def line_to_points(self, interval = 30, save_to_shapefile = True, save_to_csv = False):
		"""
			Break each line instance of the shapefile into points.
			Arguments:
				interval (float): Interval between each points (in map units, preferably meters)
				save_to_shapefile (bool): if True, save a shapefile with the new points
				save_to_csv (bool): if True, save a csv with 3 columns (line_ID, X, Y)
			Returns:
				a Pandas dataframe with 3 columns (line_ID, X, Y)
			Authors:
				B.G.
			Date:
				04/03/2019
		"""

		print("I am trying to break your line, this feature hasn't been widely tested yet, please let me know if it breaks or bugs") if (self.verbose) else 0

		# Preformatting the output
		these_X = []
		these_Y = []
		these_ID = []
		these_TC = []
		this_ID = 0

		freeST = strikethrough("free")

		# Looping through the lines
		for this_line in self.shpdf["geometry"].values:
			print("I want to break " + freeST + " line #%s" %(this_ID)) if self.verbose else 0
			# Reinitialising the distance
			this_dist = 0
			# Reiterating the interpolation while we still are on the line
			## Background about geopandas/shapely: 
			## line.length is the total length of that line;
			## interpolate returns a point at a certain distance from the origin of the line
			## xy returns weird coordinates from a point object
			while(this_dist < this_line.length):
				# Getting the point
				tC = this_line.interpolate(this_dist)
				# Saving the point
				these_TC.append(tC)
				# Extracting raw x and y
				these_X.append(tC.xy[0][0])
				these_Y.append(tC.xy[1][0])
				these_ID.append(this_ID)
				# Incrementing the interval
				this_dist += interval

			# Line is done yo, incrementing the ID
			this_ID += 1
		# Formatting output csv
		odf = pd.DataFrame({"ID":these_ID, "X": these_X, "Y": these_Y})
		
		if(save_to_shapefile):
			print("Saving to shapefile ... %s"%(self.path + self.prefix + "_line2points.shp")) if self.verbose else 0
			# Formatting shapefile output
			ogdf = odf.copy()
			ogdf["coordinates"] = these_TC
			ogdf = gpd.GeoDataFrame(ogdf, geometry = "coordinates")
			ogdf.to_file(self.path + self.prefix + "_line2points.shp")
		if(save_to_csv):
			print("Saving to shapefile ... %s"%(self.path + self.prefix + "_line2points.csv")) if self.verbose else 0
			odf.to_csv(self.path + self.prefix + "_line2points.csv", index = False)

		print("Done with breaking your lines.") if self.verbose else 0

		return odf


	def test_rectangular_window(self, distance_along_line = 200, half_width = 600, sample_step = 20, line2points = True, point = 25):
		"""
			Test function to visualise a rectangular window.
			Description to make later
		"""
		from scipy import stats

		if(line2points):
			points = self.line_to_points(interval = sample_step)
		print(points)

		# number of points to get on each side of the reference node on the line
		half_n_point_reg = round(distance_along_line / sample_step / 2)

		# number of points to get on each side of the reference node on the line
		half_n_point_side = round(half_width / sample_step / 2)

		# test function, so I am just taking the first line

		# First step is to get the base points
		Xes = points["X"].values[point - half_n_point_reg : point+ half_n_point_reg];Yes = points["Y"].values[point - half_n_point_reg: point+ half_n_point_reg]
		txmin = np.min(Xes);txmax = np.max(Xes)

		# The linear regression estimates the best m x + b for that part
		this_m, this_b, r_value, p_value, std_err = stats.linregress(Xes,Yes)

		alpha = np.arctan(this_m)

		# Now getting the resampled points
		X_A = txmin
		Y_A = this_m * X_A + this_b

		dx = sample_step * np.cos(alpha)
		dy = sample_step * np.sin(alpha)

		# sampling the new baseline
		est_Xes = [X_A]
		est_Yes = [Y_A]
		this_dist = 0
		while(this_dist < distance_along_line):
			# Calculating the new coordinates
			X_B = X_A + dx
			Y_B = Y_A + dy
			# Incrementing the cat stuffs
			this_dist += sample_step
			# Saving and moving on
			est_Xes.append(X_B)
			est_Yes.append(Y_B)
			X_A = X_B
			Y_A = Y_B

		# Alright, Now I do have the base line, I just need to find a way to get the orthogonal stuffs innit?
		list_of_orthogonal_X = []
		list_of_orthogonal_Y = []
		barking_m = -1/this_m
		barking_alpha = np.arctan(barking_m)
		barking_dx = sample_step * np.cos(barking_alpha)
		barking_dy = sample_step * np.sin(barking_alpha)
		for i in range(len(est_Xes)):
			# Just getting the bloody coordinates of the assessed point
			tx = est_Xes[i]
			ty = est_Yes[i]

			this_dist = 0
			this_barking_Xes = [tx]
			this_barking_Yes = [ty]
			increment = 1
			while(this_dist<half_width):
				this_barking_Xes.append(tx + increment * barking_dx)
				this_barking_Xes.append(tx - increment * barking_dx)
				this_barking_Yes.append(ty + increment * barking_dy)
				this_barking_Yes.append(ty - increment * barking_dy)
				this_dist += sample_step
				increment += 1
			list_of_orthogonal_X.append(this_barking_Xes)
			list_of_orthogonal_Y.append(this_barking_Yes)

		# TRUE OUTPUT
		return list_of_orthogonal_X, list_of_orthogonal_Y
		# FAKE OUTPUT FOR TESTING PURPOSES
		# return est_Xes, est_Yes



		













######## Random small functions

def strikethrough(text):
    result = ''
    for c in text:
        result = result + c + '\u0336'
    return result


"""
Quick visualisation routines for m_over_n analysis.
B.G
"""
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
import os
import lsdtopytools.quickplot_utilities as QUT
from lsdtopytools import LSDDEM

def plot_disorder_results(mydem, normalise = False, figsize = (4,3), dpi = 300, output = "save", format_figure = "png", legend = True,
 cumulative_best_fit = True):
	"""
		Plot results from simple disorder analysis (Mudd et al., 2018).
		The value represents the scatter of chi-elevation plots for each basins: High value = more scatter, low value = less scatter.
		Note that if you have basins of different area, their value is harder to compare, but you can normalise it to the maximum.
		Arguments:
			mydem: an LSDDEM object where the disorder best-fit has been calculated
			normalise (bool): False to plot the absolute disorder value, True to normalise to maximum (0-1). You want to normalise if you need to actually compare several basins as the disorder vallue is linked to the basin size.
			figsize (tuple or list of 2): width, heigth in matplotlib units (inches I guess).
			dpi (int): figure quality if saveing ouput is chosen
			output (str): "save" to save figure, "return" to return fig/axis, "show" to show.
			legend (bool): Do you need to plot the legend. True or False.
			cumulative_best_fit (bool): Add something to the plot.
		Returns:
			Depends on the output size			
		Authors:
			Boris Gailleton
		Date:
			04/01/2019
	"""
	# Preparing figure
	fig = plt.figure(figsize = figsize)
	gs = plt.GridSpec(100,100,bottom=0.15, left=0.18, right=0.87, top=0.98)
	ax1 = fig.add_subplot(gs[0:100,0:100], facecolor = "none") # Plot the movern scatter and lien

	# Getting the data
	X_movern = mydem.cppdem.get_disorder_vec_of_tested_movern()
	dict_per_basins = mydem.cppdem.get_disorder_dict()

	best_fit = np.zeros(len(X_movern))

	# Plotting it
	for key,val in dict_per_basins.items():
		tval = np.asarray(val)
		tval = tval/np.nanmax(tval) if (normalise) else tval
		# Plotting the line
		ax1.plot(X_movern, tval, lw =1, alpha = 0.8, zorder = 1, label = key)
		ax1.scatter(X_movern, tval, lw =1, s = 5, c = "k", zorder = 2, marker = "+")
		best_fit[np.argmin(tval)] += 1 # Incrementing the cumulative best-fit

	if(cumulative_best_fit):
		ax2 = ax1.twinx() # plot the cumulative best fit
		ax2.plot(X_movern, best_fit, color = "k", lw = 1, ls = "-.", zorder = 0.5)
		ax2.fill_between(X_movern,0,best_fit,lw =0, color = "k", alpha = 0.3333)
		ax2.set_ylabel("Best-fit cumulative")

	ax1.legend() if (legend) else 0
	ax1.set_xlabel(r"$\theta$ tested")
	ylab = "Disorder"
	ylab += " (norm.)" if(normalise) else ""
	ax1.set_ylabel(ylab)

	return QUT.finalise_fig(fig, ax1, output, mydem, "concavity_disorder_results", format_figure, dpi)

def plot_disorder_map(mydem ,figure_width = 4, figure_width_units = "inches", cmap = "jet", alpha_hillshade = 0.95, 
	this_fontsize = 6, alpha_catchments = 0.75,  dpi = 300, output = "save", format_figure = "png"):
	"""
		Plot a map of best fit theta using disorder method. You can adjust the following set of parameters to customise it:
		Arguments:
			mydem: the LSDDEM object. Required.
			figure_width (float): figure width in inches (Default value because of matplotlib...) or centimeters.
			figure_width_units (str): "inches" (Default) or "centimetres"
			cmap (str): the name of the colormap to use, see https://matplotlib.org/examples/color/colormaps_reference.html for list.
			alpha_hillshade (float): regulate the transparency of the background hillshade. Between 0 (transparent) and 1 (opaque).
			dpi (int): Figure quality, increase will increase the figure quality and its size ofc.
			output (str): "save" to save figure, "return" to return fig/axis, "show" to show.
			format_figure (str): the extension of the saved bitmap: "png", "jpg", "svg" reccomended. For the rest see: https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
			this_fontsize (flaot): font size of minor texts
			alpha_catchments: transparency of basin colors.
		Returns:
			Depends on parameters
		Authors:
			Boris Gailleton
		Date:
			19/12/2018
	"""

	# First creating the figure and axis
	fig = QUT.create_single_fig_from_width(mydem.ncols, mydem.nrows, width = figure_width, width_units = figure_width_units, background_color = "white")
	gs = plt.GridSpec(100,100,bottom=0.15, left=0.20, right=0.98, top=0.98)
	ax1 = fig.add_subplot(gs[0:100,0:100], facecolor = "none")

	# Plotting the figure
	ax1.imshow(mydem.get_hillshade(), extent = mydem.extent, interpolation = "none", cmap = "gray", vmin= 0, vmax = 255, alpha = alpha_hillshade)

	# Getting the catchment data
	BAS = mydem.cppdem.get_chi_basin().astype(np.float32)# switching to float32 to enable nan support for transparency
	BAS[BAS==-9999] = np.nan

	# Getting the movern data
	# Getting the data
	X_movern = mydem.cppdem.get_disorder_vec_of_tested_movern()
	dict_per_basins = mydem.cppdem.get_disorder_dict()

	best_fit = np.array(X_movern)

	# Plotting it
	for key,val in dict_per_basins.items():
		tval = np.asarray(val)
		bf = best_fit[np.argmin(tval)] # Incrementing the cumulative best-fit
		BAS[BAS == key] = bf


	cb = ax1.imshow(BAS, extent = mydem.extent, interpolation = "none", cmap = cmap, alpha = alpha_catchments)
	cax = plt.colorbar(cb )
	cax.ax.set_ylabel(r"Best fit $\theta$", fontsize = this_fontsize, rotation = -270)
	# cax.ax.get_yaxis().labelpad = -10
	cax.ax.tick_params(labelsize = this_fontsize) 
	QUT.fix_map_axis_to_kms(ax1, this_fontsize, figure_width)
	return QUT.finalise_fig(fig, ax1, output, mydem, "concavity_disorder_map", format_figure, dpi)


#!/usr/bin/env python
"""
Command-line tool to control the concavity constraining tools
Mudd et al., 2018
So far mostly testing purposes
B.G.
"""
import matplotlib;matplotlib.use("Agg")
from lsdtopytools import LSDDEM # I am telling python I will need this module to run.
from lsdtopytools import argparser_debug as AGPD # I am telling python I will need this module to run.
from lsdtopytools import quickplot as qp, quickplot_movern as qmn # We will need the plotting routines
import time as clock # Basic benchmarking
import sys # manage the argv
import numpy as np
from matplotlib import pyplot as plt

def chi_mapping_tools():
	# Here are the different parameters and their default value fr this script
	default_param = AGPD.get_common_default_param()
	default_param["map"] = False
	default_param["theta_ref"] = 0.45
	default_param["A_0"] = 1
	default_param["area_threshold"] = 500

	default_param["X"] = None
	default_param["Y"] = None

	default_param = AGPD.ingest_param(default_param, sys.argv)

	# Checking the param
	if(isinstance(default_param["X"],str)):
		X = [float(default_param["X"])]
		Y = [float(default_param["Y"])]
	else:
		X = [float(i) for i in default_param["X"]]
		Y = [float(i) for i in default_param["Y"]]


	if(default_param["help"] or len(sys.argv)==1):
		print("""
			This command-line tool run chi analysis tools from LSDTopoTools. This tool is focused on extracting chi, ksn or things like that.
			Chi -> DOI: 10.1002/esp.3302 An integral approach to bedrock river profile analysis, Perron and Royden 2013
			ksn -> (calculated from a robust statistical method based on chi): https://agupubs.onlinelibrary.wiley.com/doi/epdf/10.1002/2013JF002981, Mudd et al., 2014

			
			To use, run the script with relevant options, for example:
				lsdtt-chi-tools file=myraster.tif map X=531586 Y=6189787 theta_ref=0.55 A_0=10

			option available:
				file: name of the raster (file=name.tif)
				path: path to the file (default = current folder)
				map: generate map of chi
				theta_ref: theta ref on the main equation
				A_0: A_0 on the chi equation
				area_threshold: the threshold of pixel to initiate the river network
				X: X coordinate of the outlet (So far only single basin is supported as this is an alpha tool)
				Y: Y coordinate of the outlet (So far only single basin is supported as this is an alpha tool)
				help: if written, diplay this message. Documentation soon to be written.

			""")

		quit()


	print("Welcome to the command-line tool to plot basin-wide hypsometry stuff!")
	print("Let me first load the raster ...")
	mydem = LSDDEM(file_name = default_param["file"], path=default_param["path"], already_preprocessed = False, verbose = False)
	print("Got it. Now dealing with the depressions ...")
	mydem.PreProcessing(filling = True, carving = True, minimum_slope_for_filling = 0.0001) # Unecessary if already preprocessed of course.
	print("Done! Extracting the river network")
	mydem.ExtractRiverNetwork( method = "area_threshold", area_threshold_min = 1000)
	print("I have some rivers for you! Defining the watershed of interest...")


	mydem.DefineCatchment( method="from_XY", X_coords = X, Y_coords = Y, test_edges = False, coord_search_radius_nodes = 25, coord_threshold_stream_order = 1)
	print("I should have it now.")
	print("I got all the common info, now I am running what you want!")
	mydem.GenerateChi(theta = float(default_param["theta_ref"]) , A_0 = float(default_param["A_0"]))

	print("Producing a figure of your catchement just for you to check if their location is OK!")
	qp.plot_check_catchments(mydem)


	if default_param["map"]:
		print("I am going to plot chi maps")
		fig,ax = qp.plot_nice_topography(mydem ,figure_width = 4, figure_width_units = "inches", cmap = "gist_earth", hillshade = True, 
			alpha_hillshade = 1, color_min = None, color_max = None, dpi = 300, output = "return", format_figure = "png", fontsize_ticks =6, fontsize_label = 8, 
			hillshade_cmin = 0, hillshade_cmax = 250,colorbar = False, colorbar_label = None, colorbar_ax = None)


		normalize = matplotlib.colors.Normalize(vmin = mydem.df_base_river["drainage_area"].min(), vmax = mydem.df_base_river["drainage_area"].max())
		ax.scatter(mydem.df_base_river["x"],mydem.df_base_river["y"], s = normalize(mydem.df_base_river["drainage_area"].values), c = mydem.df_base_river["chi"], vmin = normalize(mydem.df_base_river["chi"].quantile(0.10)), vmax = normalize(mydem.df_base_river["chi"].quantile(0.90)), zorder = 5)

		plt.savefig(mydem.save_dir+mydem.prefix+"_chi_map"+"."+"png", dpi = 500)
		plt.clf()




	print("Finished!")




"""
This file contains supporting routines for the quickplot facilities.
For example, determining the right size for a figure or extracting basin outines.
It might send data to cpp at some point, I don't know but it does contains useful small routines I ended up 
using in other project so feel free to pick some in it.
B.G. 19/12/2018
"""
from matplotlib import pyplot as plt
import numpy as np


def create_single_fig_from_width(plotXextent, plotYextent,width = 4, width_units = "inches", background_color = "white"):
	"""
	Creates a figure with an adapted height from its width.
	Arguments:
		plotXextent: the size of the plotted object, to determine the best ratio. For a map, it would be n-columns and n-rows in the raster. For a scatter it would be xmax-xmin and ymax-ymin. ...
		width (float): the required width.
		width_units (str): inches (erk... but matplotlib's norm), or cm/centimetre/centimeter
		background_color (str): background color, can be any color recognised by matplotlib (e.g., HTML color code).
	Returns: 
		A matplotib figure object, hopefully correctly sized
	Authors:
		B.G - 19/12/2018
	"""
	# First I am getting the ratio between Y and X:
	ratioExtent = plotYextent/(plotXextent*1)
	# Easy, now eventually converting the units
	width = width/2.54 if width_units.lower() in ["cm","centimeters","centimeter","centimetre","centimetres"] else width
	return plt.figure(figsize = [width,ratioExtent*width], facecolor = background_color)

def fix_map_axis_to_kms(this_axis, this_fontsize,figure_width):
	"""
		Quick utility that take an axis and reduce its ticks from meters to km.
		Parameters:
			this_axis: the matplotlib axis
			this_fontsize(float): the font size for labels
			figure_width(float): figure width to adapt some text
		Returns:
			Nothing, changes stuff inside the ax
	"""
		# Fixing the ticks: kms rather than metres

	xti = this_axis.get_xticks()
	xtlab = []
	for i in range(len(xti)):
		xtlab.append(str(int(xti[i]/1000)))
	this_axis.set_xticklabels(xtlab, fontsize = this_fontsize)
	yti = this_axis.get_yticks()
	ytlab = []
	for i in range(len(yti)):
		ytlab.append(str(int(yti[i]/1000)))
	this_axis.set_yticklabels(ytlab, fontsize = this_fontsize)

	# labels:
	this_axis.set_xlabel("Easting (km)", fontsize = this_fontsize * 1.2)
	this_axis.set_ylabel("Northing (km)", fontsize = this_fontsize * 1.2)

def finalise_fig(this_fig, these_ax, output, mydem, suffix, format_figure, dpi):
	"""
		When any figure is created with the quickplot engine, user can choose between "finalising" methods:
		The figure can be saved, showed (depending on your matplotlib backend), returned, or leave as it is.
		More details in the argument description.
		Arguments:
			this_fig: The matplotlib figure object
			these_ax: the (single or list of) ax associated with the plotting. Multiple axes are for more complexe plots.
			output (str): the output method. Can be:
				"show": will show the figure with plt.show(). Can not work if your backend is not properly setted.
				"save": will save the figure to a file.
				"return": will return fig, ax for further modifications. For example if you want to add your own scatter on the top of a map, and you just want quickplot to plot the "basemap" basically.
				"Nothing": In some cases, the figure can be passed as argument from a previous manipulation to add the quickplot on top of that and get it back. Quite rare, I needed that for specific cases such as animation plotting. it's there is you need it.
			suffix (str): in case of saving output, this suffix is the name of the fig before the ".". I don't remember why I named it suffix, it made sense back in the days.
			format_figure (str): In case of saving output, this is the format of the figure. It depends on your matplotlib backend, "png" "svg" "pdf" work most of the time.
		Returns:
			Depends on the output describe above
		Authors:
			B.G.
		Date:
			11/2018 (last update on the 23/02/2019)
	"""
	# returning it eventually
	if(output == "show"):
		plt.show()
		plt.clf()
	elif(output == "save"):
		plt.savefig(mydem.save_dir+mydem.prefix+"_"+suffix+"."+format_figure, dpi = dpi)
		plt.clf()
	elif(output == "return"):
		return this_fig,these_ax
	elif(output == "nothing"):
		print("You're ouput mode is set to nothing, ain't returning stuff but modify your fig/axes in place.")
	else:
		print("FATALERROR::I did not recognise your output option. Aborting.")
		quit() 

def add_basin_outlines(mydem, fig, ax, size_outline = 1, zorder = 2, color = "k"):
	"""
		Add catchment outlines to any axis given with its associated LSDDEM object ofc.
		No need to mention that catchment needs to be extracted within the LSDDEM object beforehand.
		Arguments:
			fig: the matplotlib figure
			ax: the matplotlib ax
			size_outline: size (in matplotlib units) of the outline
			zorder: the matplotlib zorder to use
			color (any color code compatible with matplotlib): the color of the outline. Default is black
		Returns:
			Nothing, It directly acts on the figure passed in argument
		Authors:
			B.G.
		Date:
			12/2018 (last update on the 23/02/2018)
	"""
	# getting dict of perimeters
	outlines = mydem.cppdem.get_catchment_perimeter()
	# keys are x,y,elevation,basin_key
	for key,val in outlines.items():
		print(key)
		ax.scatter(val["x"], val["y"], c = color, s = size_outline, zorder = 100, lw =0)

def size_my_points(Iarray, minsize,maxsize):
	"""
		Optimising my point normalisation here. Will be coded once for all.
		Take an array (Initial Array -> Iarray) and returns a normalised version with new bounds.
		Argumetns:
			Iarray(numpy nd_array): The initial array
			minsize (float): the new minimum
			maxsize (float): the new maximum
		Returns:
			The new array
		Authors:
			B.G. and Wikipedia for the normalisation formula
		Date:
			12/2018 (last update: 23/02/2019)
	"""
	# getting min and max
	tmax = np.nanmax(Iarray)
	tmin = np.nanmin(Iarray)
	Iarray = (Iarray-tmin)/(tmax-tmin)
	# Now my data is between 0 and 1
	Iarray = (Iarray * (maxsize-minsize)) + minsize
	# 0_1 -> 0_(maxsize - minsize) -> minsize_maxsize
	return Iarray



"""
This file contains plotting routines organised in independent functions.
I don't know how I will separate these in the future, but quickplot aims to provide very basic first order plots
All the routines should leave 3 possible ouputs: save the figure, get the figure (for further modifications) and show the figure.
B.G. 2018
"""
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
import os
import lsdtopytools.quickplot_utilities as QUT
from lsdtopytools import LSDDEM

def plot_nice_topography(mydem ,figure_width = 4, figure_width_units = "inches", cmap = "gist_earth", hillshade = True, 
	alpha_hillshade = 0.45, color_min = None, color_max = None, dpi = 300, output = "save", format_figure = "png", fontsize_ticks =6, fontsize_label = 8, 
	hillshade_cmin = 0, hillshade_cmax = 250,colorbar = False, 
	fig = None, ax = None, colorbar_label = None, colorbar_ax = None):
	"""
		Plot a nice topographic map. You can adjust the following set of parameters to customise it:
		Arguments:
			mydem: the LSDDEM object. Required.
			figure_width (float): figure width in inches (Default value because of matplotlib...) or centimeters.
			figure_width_units (str): "inches" (Default) or "centimetres"
			cmap (str): the name of the colormap to use, see https://matplotlib.org/examples/color/colormaps_reference.html for list.
			hillshade (bool): Drape an hillshade on the top of the figure
			alpha_hillshade (float): regulate the transparency of the hillshade. Between 0 (transparent) and 1 (opaque).
			color_min (float): min elevation corresponding to the min color on the colormap. None = auto.
			color_max (float): see color_min and replace min by max.
			dpi (int): Figure quality, increase will increase the figure quality and its size ofc.
			output (str): "save" to save figure, "return" to return fig/axis, "show" to show.
			format_figure (str): the extension of the saved bitmap: "png", "jpg", "svg" reccomended. For the rest see: https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
			fontsize_ticks (float): the fontsize of the ticks
			fontsize_label (float): font size of the labels
			fig (matplotlib figure): Default None, in certain cases you may want to add teh basemap plot on the top of an existing. 
			ax (Matplotlib axis): Default None. if custom Figure passed as argument, you also need to pass an axis
			colorbar_label (str): label of the colorbar if activated.
			colorbar_ax (mpl axis): the axis where to place the colorbar (Default will be auto no worries)
		Returns:
			Depends on parameters
		Authors:
			Boris Gailleton
		Date:
			19/12/2018 - last update 09/03/2019
	"""

	# First creating the figure and axis
	if(fig is None):
		fig = QUT.create_single_fig_from_width(mydem.ncols, mydem.nrows, width = figure_width, width_units = figure_width_units, background_color = "white")
	gs = plt.GridSpec(100,100,bottom=0.15, left=0.20, right=0.98, top=0.98)
	if(ax is None):
		ax1 = fig.add_subplot(gs[0:100,0:100], facecolor = "none")
	else:
		ax1 = ax

	# Plotting the figure
	cb = ax1.imshow(mydem.cppdem.get_PP_raster(), extent = mydem.extent, interpolation = "none", cmap = cmap, vmin= color_min, vmax = color_max, zorder = 1)
	if( (hillshade_cmin is not None) and (hillshade_cmin is not None)):
		cb2 = ax1.imshow(mydem.get_hillshade(), extent = mydem.extent, interpolation = "none", cmap = "gray", vmin= hillshade_cmin, vmax = hillshade_cmax, alpha = alpha_hillshade, zorder = 2)
	else:
		arr = mydem.get_hillshade()
		arr[arr==-9999] = np.nan
		cb2 = ax1.imshow(arr, extent = mydem.extent, interpolation = "none", cmap = "gray", alpha = alpha_hillshade, zorder = 2)
	if(colorbar):
		if(colorbar_ax is None):
			cbar = plt.colorbar(cb)
		else:
			cbar = plt.colorbar(cb,colorbar_ax)

		cbar.ax.tick_params(labelsize=fontsize_ticks)

		if(~(colorbar_label is None)):
			cbar.ax.set_ylabel(colorbar_label)


	# Fixing the ticks: kms rather than metres
	QUT.fix_map_axis_to_kms(ax1, fontsize_ticks, figure_width)
	# labels:
	ax1.set_xlabel("Easting (km)", fontsize = fontsize_label)
	ax1.set_ylabel("Northing (km)", fontsize = fontsize_label)

	return QUT.finalise_fig(fig, ax1, output, mydem, "_topo.", format_figure, dpi)


def plot_preprocessing_diff(mydem ,figure_width = 4, figure_width_units = "inches", cmap = "RdBu_r", this_fontsize = 6,
	color_min = None, color_max = None, dpi = 300, output = "save", format_figure = "png", n_dec_lab = 5):
	"""
		Plot a differential raster between carving/breaching/filling and the original data. Useful to check preprocessing artifacts.
		Arguments:
			mydem: the LSDDEM object. Required.
			figure_width (float): figure width in inches (Default value because of matplotlib...) or centimeters.
			figure_width_units (str): "inches" (Default) or "centimetres"
			cmap (str): the name of the colormap to use, see https://matplotlib.org/examples/color/colormaps_reference.html for list.
			color_min (float): min elevation corresponding to the min color on the colormap. None = auto.
			color_max (float): see color_min and replace min by max.
			dpi (int): Figure quality, increase will increase the figure quality and its size ofc.
			output (str): "save" to save figure, "return" to return fig/axis, "show" to show.
			format_figure (str): the extension of the saved bitmap: "png", "jpg", "svg" reccomended. For the rest see: https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
			n_dec_lab (int): A bit experimental, it manage the precision of colorbar labels.
			this_fontsize (float): fontsize of the minor labels
		Returns:
			Depends on parameters
		Authors:
			Boris Gailleton
		Date:
			19/12/2018
	"""

	# First creating the figure and axis
	fig = QUT.create_single_fig_from_width(mydem.ncols, mydem.nrows, width = figure_width, width_units = figure_width_units, background_color = "white")
	gs = plt.GridSpec(100,100,bottom=0.15, left=0.18, right=0.98, top=0.98)
	ax1 = fig.add_subplot(gs[0:100,0:100], facecolor = "none")

	# Get the differential raster
	pre = mydem.cppdem.get_base_raster()
	post = mydem.cppdem.get_PP_raster()
	pre[pre==-9999] = np.nan
	post[post==-9999] = np.nan
	diff = post - pre


	#Dealing wih the colorbar
	if(color_min is None):
		this_min = np.nanmin(diff)
	else:
		this_min = color_min
	if(color_max is None):
		this_max = np.nanmax(diff)
	else:
		this_max = color_max

	# Plotting the figure
	has_been_plotted = False
	is_min_cax = False
	if(this_max>0):
		has_been_plotted = True
		tdiff = np.copy(diff)
		tdiff[diff<0] = np.nan
		cb = ax1.imshow(tdiff, extent = mydem.extent, interpolation = "none", cmap = cmap, vmin = -this_max, vmax = this_max)
		del tdiff
	if(this_min<0):
		is_min_cax = True
		has_been_plotted = True
		tdiff = np.copy(diff)
		tdiff[tdiff>0] = np.nan
		cb = ax1.imshow(tdiff, extent = mydem.extent, interpolation = "none", cmap = cmap, vmin = this_min, vmax = -this_min)
		del tdiff


	# Check
	if(has_been_plotted == False):
		print("Not a single change in the preprocessing, Aborting the plot, it will be an empty plot")
	else:
		cax = plt.colorbar(cb )
		cax.ax.set_ylabel(r"$\Delta$ elevation", fontsize = this_fontsize, rotation = -270)
		cax.ax.get_yaxis().labelpad = -10
		extr = cax.get_ticks()
		if(is_min_cax):
			cax.set_ticks([this_min,0,-this_min])
		if(is_min_cax== False):
			cax.set_ticks([-this_max,0,this_max])

		cax.set_ticklabels([str(this_min)[:n_dec_lab],0,str(this_max)[:n_dec_lab]])
		cax.ax.tick_params(labelsize = this_fontsize) 



	# Fixing the ticks: kms rather than metres
	xti = ax1.get_xticks()
	xtlab = []
	for i in range(len(xti)):
		xtlab.append(str(xti[i]/1000))
	ax1.set_xticklabels(xtlab, fontsize = this_fontsize)
	yti = ax1.get_yticks()
	ytlab = []
	for i in range(len(yti)):
		ytlab.append(str(yti[i]/1000))
	ax1.set_yticklabels(ytlab, fontsize = this_fontsize)

	# labels:
	ax1.set_xlabel("Easting (km)", fontsize = this_fontsize * 1.2)
	ax1.set_ylabel("Northing (km)", fontsize = this_fontsize * 1.2)


	# returning it eventually
	if(output == "show"):
		plt.show()
		plt.clf()
	elif(output == "save"):
		# Dealing with the particle
		if(mydem.has_been_filled_by_me and mydem.has_been_carved_by_me):
			part = "PP_diff."
		elif(mydem.has_been_carved_by_me):
			part = "only_carved_diff."
		elif(mydem.has_been_filled_by_me):
			part = "only_filled_diff."
		# Saving the figure to the disk
		plt.savefig(mydem.save_dir+mydem.prefix+part+format_figure, dpi = dpi)

	elif(output == "return"):
		return fig,ax
	else:
		print("FATALERROR::I did not recognise your output option. Aborting.")
		quit()






def plot_check_catchments(mydem ,figure_width = 4, figure_width_units = "inches", cmap = "jet", alpha_hillshade = 0.95, 
	this_fontsize = 6, alpha_catchments = 0.75,  dpi = 300, output = "save", format_figure = "png"):
	"""
		Plot to check basin selection. You can adjust the following set of parameters to customise it:
		Arguments:
			mydem: the LSDDEM object. Required.
			figure_width (float): figure width in inches (Default value because of matplotlib...) or centimeters.
			figure_width_units (str): "inches" (Default) or "centimetres"
			cmap (str): the name of the colormap to use, see https://matplotlib.org/examples/color/colormaps_reference.html for list.
			alpha_hillshade (float): regulate the transparency of the background hillshade. Between 0 (transparent) and 1 (opaque).
			dpi (int): Figure quality, increase will increase the figure quality and its size ofc.
			output (str): "save" to save figure, "return" to return fig/axis, "show" to show.
			format_figure (str): the extension of the saved bitmap: "png", "jpg", "svg" reccomended. For the rest see: https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
			this_fontsize (flaot): font size of minor texts
			alpha_catchments: transparency of basin colors.
		Returns:
			Depends on parameters
		Authors:
			Boris Gailleton
		Date:
			19/12/2018
	"""

	# First creating the figure and axis
	fig = QUT.create_single_fig_from_width(mydem.ncols, mydem.nrows, width = figure_width, width_units = figure_width_units, background_color = "white")
	gs = plt.GridSpec(100,100,bottom=0.15, left=0.20, right=0.98, top=0.98)
	ax1 = fig.add_subplot(gs[0:100,0:100], facecolor = "none")

	# Plotting the figure
	ax1.imshow(mydem.get_hillshade(), extent = mydem.extent, interpolation = "none", cmap = "gray", vmin= 0, vmax = 255, alpha = alpha_hillshade)

	# Getting the catchment data
	BAS = mydem.cppdem.get_chi_basin().astype(np.float32)# switching to float32 to enable nan support for transparency
	BAS[BAS==-9999] = np.nan
	cb = ax1.imshow(BAS, extent = mydem.extent, interpolation = "none", cmap = cmap, vmin= 0, alpha = alpha_catchments)
	cax = plt.colorbar(cb )
	cax.ax.set_ylabel("Basin Keys", fontsize = this_fontsize, rotation = -270)
	cax.ax.get_yaxis().labelpad = -10
	cax.ax.tick_params(labelsize = this_fontsize) 
	QUT.fix_map_axis_to_kms(ax1, this_fontsize, figure_width)
	return QUT.finalise_fig(fig, ax1, output, mydem, "catchment_check", format_figure, dpi)


# -*- coding: utf-8 -*-

"""
Main module, contains the base object that host all the different analysis

Authors: B.G. 18/11/2018

"""
# This module manages raster I/O operations, based on rasterio (which itself depends on GDAL)
from lsdtopytools import raster_loader as rl
from lsdtopytools import lsdtopytools_utilities as ut
# This module is the low-level interface with the c++ code: it controls the c++ LSDDEM_xtensor object
from lsdtt_xtensor_python import LSDDEM_cpp
# Numpy provides efficient numerical arrays and shitload of useful operations with it
import numpy as np
# Pandas manages dataset as tables and provides quite a lot of fast operations
import pandas as pd
# Deals with OS operations (check folders, path and stuff aye)
import os



class LSDDEM(object):
	"""
		LSDEM is the main object controlling the module. It relies on loading a georeferenced encironmont from a raster (or eventually a csv file) and manage everything you need out of that.
		This object follows a strategy of accesser/setter with the property argument from python. The advantage of such a method is to make easy it's use for End user.
		For example, one might want to totally control all the algorithms parameter himself and extract all the metrics step by step but others might only want a quick overview of their field.
		With this approach we can make sure to conditionalize the access to each component. It will also help memory management by dropping to the disk data and only load it when required.
		
	"""
	def __init__(self, path = "./", file_name = "test.tif", already_preprocessed = False, verbose = True, prefix_supp = ""):
		"""
		Description:
			Constructor for the DEM object: initiate the program and the talk with LSDTopoTools.
		Arguments:
			path (str): Path to the file. Example on Windows: "C:/Users/s1675537/Desktop/"; on linux: "/home/Boris/Desktop/"; on Mac: No idea.
			file_name (str): The name of the file with the extension, for example "cairgons.tif"
			already_preprocessed (bool): If your Raster already is ready for flow routines (ie does not need any filling, carving or filtering), turn it to True.
		Return: 
			A LSDDEM object
		Authors:
			Boris Gailleton
		Date:
			14/12/2018
		"""
		# Saving the path and file name
		self.path = path
		self.file_name = file_name
		# Getting the prefix for saving all your files:
		self.prefix = str("".join(self.file_name.split(".")[:-1])) # -> prefix = base name without extention
		self.prefix += prefix_supp
		# Do you want me to talk? TODO: find a solution to catch and maybe stop the cout statements
		self.verbose = verbose

		#Alright getting the dem:
		print("Loading the raster from file: %s%s"%(self.path,self.file_name) ) if(self.verbose) else 0
		temp_loaded = rl.load_raster(path+file_name)
		## DEM extents and dimensions
		self.extent = temp_loaded["extent"]
		self.x_min = temp_loaded["x_min"]
		self.x_max = temp_loaded["x_max"]
		self.y_min = temp_loaded["y_min"]
		self.y_max = temp_loaded["y_max"]
		self.resolution = temp_loaded["res"]
		self.crs = temp_loaded["crs"]
		## Redirecting no data values to a single one
		ndt_ls_temp = temp_loaded["nodata"]
		# Recasting No Data to -9999. LSDTopoTools cannot deal with multiple no data. Although it would be simple, it would add useless complexity. We are not software developers.
		print("I am recasting your nodata values to -9999 (standard LSDTT)") if(self.verbose) else 0
		temp_loaded["array"][np.isin(temp_loaded["array"],ndt_ls_temp)] = -9999 # Recasting
		self.nodata = -9999 # Saving
		# nrows and ncols are the raster indexes dimensions
		self.nrows = temp_loaded["nrows"]
		self.ncols = temp_loaded["ncols"]

		# Other check
		self.check_if_preprocessed = False
		self.ready_for_flow_routines = False
		self.check_catchment_defined = False
		self.check_river_readiness = False
		self.check_flow_routines = False
		self.ksn_extracted = False
		self.check_chi_gen = False
		self.knickpoint_extracted = False

		self.has_been_carved_by_me = False
		self.has_been_filled_by_me = False

		# Initialaising empty DF
		self.df_ksn = None
		self.df_knickpoint = None


		print("Alright, let me summon control upon the c++ code ...") if(self.verbose) else 0
		self.cppdem = LSDDEM_cpp(self.nrows, self.ncols, self.x_min, self.y_min, self.resolution, self.nodata, temp_loaded["array"])
		del temp_loaded # realising memory (I hope at least ahah)
		print("Got it.") if(self.verbose) else 0

		if(already_preprocessed):
			print("WARNING: you are telling me that the raster is already preprocessed. You mean either you don't need flow routine (e.g., slope, curvature, roughness,... calculations) or you already made sure you got rid of your depressions.") if(self.verbose) else 0
			self.cppdem.is_already_preprocessed()
			self.ready_for_flow_routines = True
			self.check_if_preprocessed = True


		if not os.path.exists(path+self.prefix+"/"):
		    os.makedirs(path+self.prefix+"/")

		self.save_dir = path+self.prefix+"/"
		self.hdf_name = self.save_dir+self.prefix+"_dataset.hdf"

		exists = os.path.isfile(self.save_dir+self.prefix+"_dataset.hdf")
		if not exists:
			df = pd.DataFrame({"created": [True]})
			ut.save_to_database(self.hdf_name,"Init", df)



		print("lsdtopytools is now ready to roll!") if(self.verbose) else 0

	def PreProcessing(self, filling = True, carving = True, minimum_slope_for_filling = 0.0001):
		"""
		Description:
			Any dem is noisy at a certain extent. To process flow routines, this function proposes algorithm to preprocess dem cells and ensure there is no vicious pit blocking fow path.
			Filling is currently using Wang et al., 2006 and carving is using Lindsay et al., 2016.
			Filling makes sure that a minimum slope is induxed to each cells, carving breaches the pit to let the flow go.
		Arguments:
			filling (bool): do you want to fill?
			carving (bool): Wanna carve mate?
			minimum_slope_for_filling (float): Minimum gradient to induce between each cells when filling your dem.
		Return:
			Nothing, calculate the PPRaster in the cpp object
		Authors:
			Boris Gailleton
		Date:
			14/12/2018
		"""
		# Calling the cpp interface:
		print("Carving: implementation of Lindsay (2016) DOI: 10.1002/hyp.10648") if(self.verbose and carving) else 0
		print("Filling: implementation of Wang and Liu (2006): https://doi.org/10.1080/13658810500433453") if(self.verbose and filling) else 0
		print("Processing...") if(self.verbose) else 0
		self.cppdem.PreProcessing(carving, filling, minimum_slope_for_filling)
		print("DEM ready for flow routines!") if(self.verbose) else 0
		self.check_if_preprocessed = True
		self.ready_for_flow_routines = True
		self.has_been_filled_by_me = True  if(filling) else False
		self.has_been_carved_by_me = True if(carving) else False


	def CommonFlowRoutines(self):
		"""
		Description:
			Most of the algorithms need a common base of flow routines to be preprocessed: flow accumulation, drainage area, ... as well as 
			hidden LSDTT required to run analysis. This takes care of that for you.
		Arguments:
		Returns:
			Nothing but initiate a bunch of attributes in the c++ object
		Authors:
			Boris Gailleton
		Date:
			14/12/2018
		"""
		if(self.check_if_preprocessed != True):
			print("WARNING!! You did not preprocessed the dem, I am defaulting it. Read the doc about preprocessing for more information.")
			self.PreProcessing()
		self.check_flow_routines = True
		print("Processing common flow routines...") if(self.verbose) else 0
		self.cppdem.calculate_FlowInfo()
		print("Done!") if(self.verbose) else 0

	def ExtractRiverNetwork(self, method = "area_threshold", area_threshold_min = 1000):
		"""
		Description:
			Extract river network from sources. Several methods are available to detect source locations that are then used to create the rivers.
			Available agorithms:
				minimum area threshold: If the exact location of the channel heads is not important or if you are working in relatively low precision raster (e.g.SRTM), this method really quickly initiate channels where a minimum of flow accumulation is reached.
				TODO: wiener and DREICH
		Arguments:
			method (str): Name of the method to use. Can be area_threshold, wiener or dreicht
			area_threshold_min (int): in case you chose the area_threshold method, it determines the minimum number of pixels for extracting 
		Returns:
			Nothing, generates the channel network
		Authors:
			Boris Gailleton
		Date:
			14/12/2018
		"""
		if(self.check_if_preprocessed != True):
			print("WARNING!! You did not preprocessed the dem, I am defaulting it. Read the doc about preprocessing for more information.")
			self.PreProcessing()
		if(self.check_flow_routines != True):
			print("WARNING!! You did not calculate the flow routines, let me do it first as this is required for extracting the river network.")
			self.CommonFlowRoutines()
		self.check_river_readiness = True
		if(method=="area_threshold"):
			self.cppdem.calculate_channel_heads("min_contributing_pixels", area_threshold_min)
		else:
			print("Not Implemented yet! sns.")
		self.cppdem.calculate_juctionnetwork()

	def DefineCatchment(self, method="min_area_pixels", test_edges = False, min_area_pixels = 1000, X_coords = [], Y_coords = [], coord_search_radius_nodes = 30, coord_threshold_stream_order = 3):
		"""
		Description:
			Define the catchments of interest from conditions. DISCLAIMER: work in progress, can be buggy.
			Safest option (if possible) is a list of lat-lon or XY.
		Arguments:
			method(str): name of the method to use. Can be: 
				"min_area_pixels": keep all watersheds made of min_area_pixels or more.
				"from_XY": snap location to XY coordinates
				TODO: the rest of the methods
			test_edges (bool): Ignore uncomplete catchment
			min_area_pixels (int): number of pixels to define catchments with the min_area_pixels method.
			X_coord (list or numpy array): X coordinates for method "from_XY"
			Y_coord (list or numpy array): Y coordinates for method "from_XY"
			coord_search_radius_nodes (int): radius of search around the given coordinate;
			coord_threshold_stream_order (int): minimum or maximum stream order (e.g., do you want to snap to large or small river. or even moderate, let's go crazy.)
		Returns:
			Nothing but defines a set of catchment required for some analysis (e.g., movern, knickpoints or ksn)
		Authors:
			Boris Gailleton
		Date:
			14/12/2018
		"""
		# I am first proceeding to routines check: Have you processed everything required?
		if(self.check_if_preprocessed != True):
			print("WARNING!! You did not preprocessed the dem, I am defaulting it. Read the doc about preprocessing for more information.")
			self.PreProcessing()
		if(self.check_flow_routines != True):
			print("WARNING!! You did not calculate the flow routines, let me do it first as this is required for extracting the river network.")
			self.CommonFlowRoutines()
		if(self.check_river_readiness != True):
			print("WARNING!! You did not processed the river network, let me do it first as this is required for defining the catchment.")
			self.ExtractRiverNetwork()
		self.check_catchment_defined = True
		if(method == "min_area_pixels"):
			self.cppdem.calculate_outlets_locations_from_minimum_size(min_area_pixels, test_edges, False)
		elif(method == "from_XY"):
			self.cppdem.calculate_outlets_locations_from_xy(np.array(X_coords), np.array(Y_coords), coord_search_radius_nodes, coord_threshold_stream_order, test_edges, False)

		else:
			print("Not done yet, work in progress on that")
			quit()

	def GenerateChi(self, theta=0.45,A_0 = 1):
		"""
		Generates Chi coordinates. This is needed for quite a lot of routines, and effectively also calculates quite a lot of element in the river network (node ordering and other similar thing).
		Chi coordinates is details in Perron and Royden (2013) -> DOI: 10.1002/esp.3302
		It needs preprocessing, flow routines, Extraction of river network and the definition of catchment of interest.
		Arguments:
			theta(float): concavity of profiles (m/n in Stream power like laws, theta in Flint's law). This is a really important parameter to constrain!! -> Mudd et al., 2018
			A_0(float): Reference concavity area. It is important to set it to 1 if you want to get ksn from chi coordinate (see. Mudd et al., 2014 for details) -> 10.1002/2013JF002981
		returns:
			Nothing, but calculates chi ccoordinate.
		Authors:
			B.G
		Date:
			12/2018
		"""
		# pre-checkings
		if(self.check_if_preprocessed != True):
			print("WARNING!! You did not preprocessed the dem, I am defaulting it. Read the doc about preprocessing for more information.")
			self.PreProcessing()
		if(self.check_flow_routines != True):
			print("WARNING!! You did not calculate the flow routines, let me do it first as this is required generating chi.")
			self.CommonFlowRoutines()
		if(self.check_river_readiness != True):
			print("WARNING!! You did not processed the river network, let me do it first as this is required for generating chi.")
			self.ExtractRiverNetwork()
		if(self.check_catchment_defined != True):
			print("WARNING!! You did not defined any catchment, let me do it first as this is required for generating chi.")
			self.DefineCatchment()
		self.check_chi_gen = True
		self.m_over_n = theta # Idk but we might need that at some point eventually
		self.cppdem.generate_chi(theta, A_0)
		# Trying to get the chi coordinate out of that
		D1 = self.cppdem.get_int_ksn_data()
		D2 = self.cppdem.get_float_ksn_data()
		Dict_of_ksn = {}; Dict_of_ksn.update(D1); Dict_of_ksn.update(D2); del D1; del D2;
		self.df_base_river = pd.DataFrame( { "basin_key": Dict_of_ksn["basin_key"], "col": Dict_of_ksn["col"], "nodeID": Dict_of_ksn["nodeID"],
		"row": Dict_of_ksn["row"], "source_key": Dict_of_ksn["source_key"], "chi": Dict_of_ksn["chi"], "drainage_area":  Dict_of_ksn["drainage_area"],
		"elevation":  Dict_of_ksn["elevation"], "flow_distance": Dict_of_ksn["flow_distance"], "x":  Dict_of_ksn["x"], "y":  Dict_of_ksn["y"] } )

	def ksn_MuddEtAl2014(self, target_nodes=70, n_iterations=60, skip=1, minimum_segment_length=10, sigma=2,  nthreads = 1, reload_if_same = False):
		"""
		Calculates ksn with Mudd et al., 2014 algorithm. Design to calculate ksn with a strong statistical method. More robust than SA plots or basic chi-z linear regression
		Arguments:
			target_nodes(int): Full details in paper, basically low values creates shorter segments (fine grain calculation) and higher values larger segments (large-scale trends). Values higher than 100 will be extremely slow down the algorithm.
			n_iterations(int): Full details in paper, the algorithm sample random node combinations and this parameters controls the number of iterations to get that. More iterations = less noise sensitive.
			skip(int): Full details in paper, recommended between 1-4. Higher values sample less adjacent nodes.
			minimum_segment_length(int): Do not change (debugging purposes, I shall remove that actually).
			sigma(int): Full details in paper, recommended not to change.
			nthreads (int): Experimental, leave to 1. Not ready for multithreading, will segfault.
			reload_if_same (bool): Experimental, the logic will probably change, but it uses pandas and pytables habilities to very efficiently save and reload what has already been calculated. It checks if is safe to reload or not
		Retuns: 
			Returns a string stating "generated" if recalculated or "reloaded" if data has simply been reloaded
		"""
		if(self.check_if_preprocessed != True):
			print("WARNING!! You did not preprocessed the dem, I am defaulting it. Read the doc about preprocessing for more information.")
			self.PreProcessing()
		if(self.check_flow_routines != True):
			print("WARNING!! You did not calculate the flow routines, let me do it first as this is required generating chi.")
			self.CommonFlowRoutines()
		if(self.check_river_readiness != True):
			print("WARNING!! You did not processed the river network, let me do it first as this is required for generating chi.")
			self.ExtractRiverNetwork()
		if(self.check_catchment_defined != True):
			print("WARNING!! You did not defined any catchment, let me do it first as this is required for generating chi.")
			self.DefineCatchment()
		if(self.check_chi_gen != True):
			print("WARNING!! You did not generated chi, I am defaulting it AND THAT'S BAD! I AM ASSUMING A DEFAUT CONCAVITY TO 0.45 WHICH IS NOT NECESSARILY THE CASE!!!.")
			self.GenerateChi()

		# First checking if you just want to reload
		if(reload_if_same):
			print("I am just checking if I can reload data safely")
			test_dict = {"target_nodes":target_nodes, "n_iterations":n_iterations, "skip":skip, "minimum_segment_length":minimum_segment_length, "sigma":sigma}
			tdf, met = ut.load_from_database(self.hdf_name,"ksn_MuddEtAl2014")
			is_same = True
			for key,val in test_dict.items():
				if(val != met[key]):
					is_same = False
			if(is_same):
				self.df_ksn = tdf
				print("Successfuly reloaded the file")
				self.ksn_extracted = True
				## This return statment end the function and will not recalculate ksn
				return "reloaded"
			else:
				print("Failed to safely reload, one of your paramter has changed from the calculation.")

		self.ksn_extracted = True
		if(nthreads>1):
			print("WARNING: Experimental multithreading on ksn extraction. Can (i) crash (works 99/100 of the time on Linux and 50/100 of the time on windows, because windows hates developers, seriously) or (ii) produce weird source/basin numbering (e.g. not always constant).")
		self.cppdem.generate_ksn(target_nodes, n_iterations, skip, minimum_segment_length, sigma, nthreads)
		D1 = self.cppdem.get_int_ksn_data()
		D2 = self.cppdem.get_float_ksn_data()
		Dict_of_ksn = {}; Dict_of_ksn.update(D1); Dict_of_ksn.update(D2); del D1; del D2;
		self.df_ksn = pd.DataFrame(Dict_of_ksn)
		print("I have generated ksn for the specified region!")
		print("Let me just save the result to the hdf5 file to keep track")
		ut.save_to_database(self.hdf_name, "ksn_MuddEtAl2014" , self.df_ksn, {"target_nodes":target_nodes, "n_iterations":n_iterations, "skip":skip, "minimum_segment_length":minimum_segment_length, "sigma":sigma, "shape": self.df_ksn.shape})
		return "generated"

	def knickpoint_extraction(self,lambda_TVD = "auto", combining_window = 30, window_stepped = 80, n_std_dev = 7 ):
		"""
			Runs the knickpoint extraction analysis after checking that all the river extraction routines have been done.
			The method is based on segmentation of river profile is Chi-elevation space and is described in Gailleton et al., 2019 -> https://doi.org/10.5194/esurf-7-211-2019
			It requires the calculation of chi and ksn beforehand (see ksn_MuddEtAl2014), using their default parameters is not optimal and need to be contrained!! See paper for more info.
			Arguments:
				lambda_TVD (float or str): Define the lambda parameter that reduces delta-ksn noise to extract the main break/jump in slope. Highly depends on theta (m/n), the river concavity. "auto" (default) will automatically cast lambda (see Supplementary material of the main manuscript).
				combining_window (int): Define a window that combine really close knickpoints within a certain amount of node.
				window_stepped (int): define a window that tests the current "step" compare to neighbouring nodes and detect a potential waterfall.
				n_std_dev (float): Defnie how anomalic the step needs to be compared to meighbouring ones to be a step knickpoint.
			Returns:
				Nothing, but generate the attribute df_knickpoint, a pandas dataframe containing knickpoint location.
			Authors:
				B.G.
			Date:
				11/2018 (last update: 23/02/2019)
		"""
		# First I am checking that required analysis have been done.
		# Unskippable warning if not as it is really important to be aware of that
		if(self.check_if_preprocessed != True):
			print("WARNING!! You did not preprocessed the dem, I am defaulting it. Read the doc about preprocessing for more information.")
			self.PreProcessing()
		if(self.check_flow_routines != True):
			print("WARNING!! You did not calculate the flow routines, let me do it first as this is required generating chi.")
			self.CommonFlowRoutines()
		if(self.check_river_readiness != True):
			print("WARNING!! You did not processed the river network, let me do it first as this is required for generating chi.")
			self.ExtractRiverNetwork()
		if(self.check_catchment_defined != True):
			print("WARNING!! You did not defined any catchment, let me do it first as this is required for generating chi.")
			self.DefineCatchment()
		if(self.check_chi_gen != True):
			print("WARNING!! You did not generated chi, I am defaulting it AND THAT'S BAD! I AM ASSUMING A DEFAUT CONCAVITY TO 0.45 WHICH IS NOT NECESSARILY THE CASE!!!. IT WILL HAVE A SIGNIFICANT IMPACT ON KNICKPOINT EXTRACTION AND KSN.")
			self.GenerateChi()
		if(self.check_chi_gen != True):
			print("WARNING!! You did not generated ksn, I am defaulting it: It relies on many parameters and you probably want to have a look on the documentation for that.")
			self.ksn_MuddEtAl2014()

		# Alright let's do it!
		## Just keeping track that this has been done
		self.knickpoint_extracted = True
		# Taking care of lambda
		if(isinstance(lambda_TVD,str)):
			lambda_TVD = -1
		if(lambda_TVD == -1):
			if(self.m_over_n <= 0.15):
				lambda_TVD = 0.3
			elif(self.m_over_n <= 0.2):
				lambda_TVD = 0.5
			elif(self.m_over_n <= 0.3):
				lambda_TVD = 2
			elif(self.m_over_n <= 0.35):
				lambda_TVD = 3
			elif(self.m_over_n <= 0.4):
				lambda_TVD = 5
			elif(self.m_over_n <= 0.45):
				lambda_TVD = 10
			elif(self.m_over_n <= 0.5):
				lambda_TVD = 20
			elif(self.m_over_n <= 0.55):
				lambda_TVD = 40
			elif(self.m_over_n <= 0.6):
				lambda_TVD = 100
			elif(self.m_over_n <= 0.65):
				lambda_TVD = 200
			elif(self.m_over_n <= 0.7):
				lambda_TVD = 300
			elif(self.m_over_n <= 0.75):
				lambda_TVD = 500
			elif(self.m_over_n <= 0.80):
				lambda_TVD = 1000
			elif(self.m_over_n <= 0.85):
				lambda_TVD = 2000
			elif(self.m_over_n <= 0.90):
				lambda_TVD = 5000
			elif(self.m_over_n <= 0.95):
				lambda_TVD = 10000

		# Summoning the c++ code
		# the first 2 and the 30 in the middle are artifacts of former tries and necesarry calculation. We kind of need it but they don't impact the calculation therefore can be fixed.
		self.cppdem.detect_knickpoint_locations( 2, lambda_TVD, 30, window_stepped, n_std_dev, combining_window);
		# Getting the results!!
		## Basically what is happening there is that the c++ stores the results per type. I need to gather it and combine it in python. voila voila. 
		### Getting Integer data (source key, basin, keym node index, ...)
		D1 = self.cppdem.get_int_knickpoint_data()
		### Getting floating point data (break in slope, chi, DA, ...)
		D2 = self.cppdem.get_float_knickpoint_data()
		### Combining the dict into the dataframe
		Dict_of_kp = {}; Dict_of_kp.update(D1); Dict_of_kp.update(D2);
		self.df_knickpoint = pd.DataFrame(Dict_of_kp)
		# the trimmed knickpoint dataset is first set to pointing to the raw results. It will change if you trim it to keep track of the original results.
		self.df_knickpoint_selecao = self.df_knickpoint
		# Done
		print("I got your knickpoints!") if(self.verbose) else 0

	def trim_knickpoint_dataset(self, method = "third_quartile", select_river = [], select_basin = [], trimmer = {}):
		"""
			This function generates a trimmed Dataset of knickpoint. It proposes several methods to select the rivers or basins to focus on and a bunch of quantitative way to select the main entities.
			Once a dataset is generated by the core agorithm, it needs to be trimmed as the algorithm objectively pick any knickpoint. This software however is not judgmental and won't determine the "important" knickpoint from the
			"less significant" ones. This function is where the magic happens!
			Arguments:
				method(str): "third_quartile" (delfault) will only keep the knickpoints above the third quartile of their categorie, "percentile_from_dict": choose your percentile for each categorie and "val_from_dict": choose directly the cut-off value.
				select_river (list): list of source keys to generate. Default is empty list (-> all river selected).
				select_basin (list): list of basin keys to generate. Default is empty list (-> all basin selected).
				trimmer (dict): dictionary stroring the trimming/thinning option. It contains the following keys, you need to construct and feed that function with it:
					trimmer["percentile_cut_positive_slope_break"] = ... -> Cut off for positive slope-break percentile method (percentile between 0 and 100, e.g., 12 is the 12th percentile and all the knickpoints above that woud be conversed)
					trimmer["percentile_cut_negative_slope_break"] = ... -> Cut off for negative slope-break percentile method (percentile between 0 and 100, e.g., 12 is the 12th percentile and all the knickpoints above that woud be conversed)
					trimmer["percentile_cut_step"] = ... -> Cut off for vertical step percentile method (percentile between 0 and 100, e.g., 12 is the 12th percentile and all the knickpoints above that woud be conversed)
					trimmer["value_cut_positive_slope_break"] = ... -> Cut off for positive slope-break (absolute value -> all the knickpoint above that value will be conserved)
					trimmer["value_cut_negative_slope_break"] = ... -> Cut off for positive slope-break (absolute value -> all the knickpoint above that value will be conserved)
					trimmer["value_cut_step"] = ... -> Cut off for positive slope-break (absolute value -> all the knickpoint above that value will be conserved)
			Returns:
				Nothing, but feed the attribute df_knickpoint_selecao containing the knickpoints
			Authors:
				B.G.
			Date:
				11/02/2019 (last update: 23/02/2019 by B.G.)
		""" 

		# First, trimming the river and basin keys on both dataset. Because that is what ppl want right?
		## But before I am just saving the ultimate raw data just in case
		self.df_knickpoint_Raw_AF = self.df_knickpoint.copy()
		# Actual trimming
		if(len(select_river) > 0):
			self.df_knickpoint = self.df_knickpoint[self.df_knickpoint["source_key"].isin(select_river)]
		if(len(select_basin) > 0):
			self.df_knickpoint = self.df_knickpoint[self.df_knickpoint["basin_key"].isin(basin_key)]
		# pointing around Lolz
		self.df_knickpoint_selecao = self.df_knickpoint

		# basically the trimming function takes


		


	def get_hillshade(self,altitude = 45, angle = 315, z_exageration = 1):
		"""
			Generates a hillshade from the topography. Basically simulate a sun lightening on a landscape.
			Arguments:
				altitude (float): altitude of lightening
				angle (float): Angle of lightening
				z_exageration (float): guess what? regulate the transparency. No, just kiding, exagerate the relief.
			Returns:
				a numpy 2D array, same dimensions than the topo raster, containing the hillshade values.
			Authors:
				B.G.
			Date:
				19/12/2018
		"""
		# SImply:
		return self.cppdem.get_hillshade(altitude,angle,z_exageration)


	def get_polyfit_metrics(self, window_radius = 30, average_elevation = False, slope = True, aspect = False, curvature = False, planform_curvature = False, profile_curvature = False, tangential_curvature = False, TSP = False, save_to_rast = True):
		"""
		Calculates and returns polyfit metrics calculated by fitting a window through a certain radius.
		Arguments:
			window_radius (float): radius in meter of the fitted window
			average_elevation (bool): Do you want the resampled DEM from windows fitting process
			slope (bool): Activate slope calculation
			aspect (bool): Activate the aspect calculation
			curvature (bool): Activate the curvature calculation
			planform_curvature (bool): Activate the planform curvature calculation
			profile_curvature (bool): Activate the profile curvature calculation
			tangential_curvature (bool): Activate the tangential curvature calculation
			TSP (bool): Activate the TSP calculation
			save_to_rast (bool): If activated, save output to tif file
		Returns:
			Dictionnary of 2D numpy arrays witht he same extent than mother raster. Key is parameter and value is the raster
		Authors:
			B.G
		Date:
			01/2019
		"""
		# preparing ouputs
		output = {}
		selecao = []
		names = ["average_elevation","slope","aspect","curvature","planform_curvature","profile_curvature","tangential_curvature","TSP"]
		# Preparing the binding
		selecao.append(1) if(average_elevation) else selecao.append(0);selecao.append(1) if(slope) else selecao.append(0);selecao.append(1) if(aspect) else selecao.append(0);selecao.append(1) if(curvature) else selecao.append(0);selecao.append(1) if(planform_curvature) else selecao.append(0);selecao.append(1) if(profile_curvature) else selecao.append(0);selecao.append(1) if(tangential_curvature) else selecao.append(0);selecao.append(1) if(TSP) else selecao.append(0);
		# Run the code
		list_of_getter = self.cppdem.get_polyfit_on_topo(float(window_radius), selecao)
		# sorting the output
		for i in range(len(selecao)):
			if(selecao[i] == 1):
				output[names[i]] = list_of_getter[i]
				if(save_to_rast):
					print(self.save_dir+names[i]+"_%s.tif"%(window_radius))
					rl.save_raster(list_of_getter[i], self.x_min, self.x_max, self.y_max,self.y_min, self.resolution,self.crs,self.save_dir+names[i]+"_%s.tif"%(window_radius), fmt = 'GTIFF')
			else:
				output[names[i]] = "Not calculated"


		return output



# End of file


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['Click>=6.0','numpy','pandas','rasterio',"scipy",'h5py','numba','lsdtt_xtensor_python']

setup_requirements = [ ]

test_requirements = [ ]

setup(
    author="Boris Gailleton",
    author_email='b.gailleton@sms.ed.ac.uk',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="High-level interface to LSDTopoTools in pyhton. Include as much tools as we can port in python, for the rest see our docker solutions",
    entry_points={
        'console_scripts': [
            'lsdtt-concavity-tools=scripts_for_lsdtopytools.lsdtt_concavity_tools:main_concavity',
            'lsdtt-hypsometry=scripts_for_lsdtopytools.lsdtt_hypsometry:hypsometry_tools_general',
            'lsdtt-depressions=scripts_for_lsdtopytools.lsdtt_basic_operations:PreProcess',
            'lsdtt-polyfits=scripts_for_lsdtopytools.lsdtt_basic_operations:Polyfit_Metrics',
            'lsdtt-chi-tools=scripts_for_lsdtopytools.lsdtt_chi_ksn_knickoint_tools:chi_mapping_tools'
        ],
    },
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='lsdtopytools',
    name='lsdtopytools',
    packages=find_packages(include=['lsdtopytools','scripts_for_lsdtopytools']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/bgailleton/lsdtopytools',
    version='0.0.3.4',
    zip_safe=False,
    extras_require = {
        'shapefile_tools':  ["fiona","shapely","geopandas"]
        }
)



"""
Quick scripts to deal with argument parsing for command-line tools.
Why not using argparse or click?
We will have a pretty basic but specific use of argparse and I prefer to debug it internally.
argparse is great, until you have a bug. I had hard time debugging some specific argparse. 
Why am I saying all that again?
Anyway.
B.G.
"""


def get_common_default_param():
	"""
	This function returns a dictionnary of common parameter that all command-line tools will need:
	For example, file name, path , ...
	No arguments.
	Returns:
		The base dictionnary of arguments
	Authors:
		B.G.
	Date:
		01/2019 (last update: 23/02/2019)
	"""
	compam = {}
	compam["file"]="WAWater.bil" # name of our test file.
	compam["path"]="./" # Default is current path
	compam["help"] = False

	return compam

def ingest_param(default_dict, params):
	"""
	Ingest the param and feed the dictionnary. it parses boolean type (just the name of an argument), basic input (e.g.,arg45=621) or comma-separated-values (e.g., cats=henri,garfield,Sigmund)
	Arguments:
		default_dict (python dict): dict containing all the default value for the parameter of that specific analysis
		params (list from sys.argv): the param ran on the command-line tool: 
	Authors:
		B.G.
	Date:
		01/2019 (last update: 23/02/2019)
	"""
	# Let's check all the input params
	for i in range(1,len(params)):
		arg = str(params[i])
		#Checking if this is a bool activator or a argument-bearing param
		if(len(arg.split("=")) == 1):
			# Bool
			default_dict[arg.lower()] = True
		elif( "," in arg.split("=")[1]):
			# In that case this argument is a list
			default_dict[arg.split("=")[0]] = arg.split("=")[1].split(",")

		else:
			# ADD HERE ELIF FOR OTHER CASES, for example comma separated lists of basin key
			default_dict[arg.split("=")[0]] = arg.split("=")[1]

	# Done
	return default_dict



"""
Object wrapping muddpile, the Parsimonious Integrated Landscape Evolution landscape Evoution Model 
See https://lsdtopotools.github.io/LSDTT_documentation/LSDTT_MuddPILE.html for more details.
Still quite in development and early testing for my Carpathians paper.
Will document when it will be worth it and stable.
B.G.
"""
import pandas as pd
import numpy as np
try:
    from lsdtt_xtensor_python import muddpyle_cpp
    installed = True
except ImportError:
    installed = False
import os, glob, sys
from lsdtopytools import raster_loader as rl, LSDDEM, quickplot as qp



class MuddPyle(object):
    """docstring for MuddPyle"""
    def __init__(self, mode = "initial_surface", verbose = True, save_dir = "./", prefix = "BORIS",
        IS_nrows = 250, IS_ncols = 400, IS_resolution = 30, IS_diamond_square_order = 2, IS_diamond_square_relief = 50, IS_Parabola_relief = 50, IS_roughness_reflief = 5, IS_nitnlh = 0,
        m=1.1, n=1.9,Ke = 1e-7, Diffusion = 0.002, Sc = 1., uplift_field = 0.0005,
        ):
        """
           Constructor for MuddPyle object 
        """

        self.verbose = verbose
        self.save_dir = save_dir
        self.prefix = prefix
        if(mode=="initial_surface"):
            self.cppmod = muddpyle_cpp(self.verbose)
            self.cppmod.set_save_path_name(save_dir, prefix)
            self.nrows = IS_nrows
            self.ncols = IS_ncols
            self.resolution = IS_resolution
            self.cppmod.initialise_model_kimberlite(self.nrows,self.ncols, IS_resolution, IS_diamond_square_order, IS_diamond_square_relief, IS_Parabola_relief, IS_roughness_reflief, IS_nitnlh)
           

        # Checking the inputs
        tke = Ke[0,0] if isinstance(Ke, np.ndarray) else Ke

        #TEMP RECASTING
        uplift_field = np.zeros((self.nrows,self.ncols)) + uplift_field if(~(isinstance(uplift_field,np.ndarray))) else 0
        Ke = np.zeros((self.nrows,self.ncols)) + Ke if(~(isinstance(Ke,np.ndarray))) else 0

        # Default features
        self.cppmod.set_global_default_param(Diffusion, Sc, tke)
        ## Dealing with erodibility type
        if isinstance(Ke, np.ndarray):
            self.erodibility_dtype = "variable" 
        else:
            self.erodibility_dtype = "fixed"

        ## Dealing with erodibility type
        if isinstance(Ke, np.ndarray):
            self.uplift_dtype = "variable" 
        else:
            print("Muddpyle is under construction and only support array of uplift so far.")
            print("See https://lsdtopotools.github.io/LSDTT_documentation/LSDTT_MuddPILE.html for direct use of MuddPILE")
            self.uplift_dtype = "fixed"
            quit()


        self.uplift_field = uplift_field
        self.erodibility_field = Ke

        # Dealing with the cpp object!! important
        self.cppmod.set_uplift_raster(self.uplift_field)
        self.cppmod.set_K_raster(self.erodibility_field)
        ## concavity
        self.cppmod.set_SPL_exponents(m,n)
        ## Leave that alone FROMAGE
        self.cppmod.set_uplift_mode(0)
        ## BOUNDATY CONDITION FIXED FOR THE MOMENT
        self.cppmod.set_boundary_conditions(["b","p","b","p"])
        self.current_frame = 1
        self.current_time = 0



        # Creating the directory if not existing
        directory = os.path.dirname(self.save_dir)
        if not os.path.exists(directory):
            os.makedirs(directory)



    def run_model(self,dt = 250, time = 250000, save_step = 500, fluvial = True, hillslope = False):
        import math

        self.cppmod.deal_with_time(dt, dt, self.current_time+time,0)    
        self.cppmod.set_print_interval_in_year(save_step)
        # Run the model
        self.cppmod.run_model(fluvial, hillslope,True, self.current_frame, False)
        # updtae the save charac n stuffs
        self.current_frame += math.floor(time/dt)+1
        self.current_time += time

    def update_Ke(self, new_Ke = 1e-5):

        if(~(isinstance(new_Ke,np.ndarray))):
            new_Ke = np.zeros((self.nrows,self.ncols)) + new_Ke

        self.erodibility_field = new_Ke
        self.cppmod.set_K_raster(self.erodibility_field)

    def update_Uplift(self, new_U = 0.0001):
        print("TESTTTTTTTTTT")
        if(~(isinstance(new_U,np.ndarray))):
            new_U = np.zeros((self.nrows,self.ncols)) + new_U

        self.uplift_field = new_U
        self.cppmod.set_uplift_raster(self.uplift_field)

    def plot_anim(self, fps = 10, rename_suffix = "", dpi = 300, HS_alpha = 0.75):

        from matplotlib.animation import FFMpegWriter
        from matplotlib import pyplot as plt

        # Beware of the nested function
        def plotty(direc,fname, ret = False, figt = None, dpi = 300, HS_alpha = 0.85):
            tdem = LSDDEM(path = direc, file_name = fname,already_preprocessed = True)
            mx = np.percentile(tdem.cppdem.get_PP_raster(), 90)
            if(ret):
                return qp.plot_nice_topography(tdem ,figure_width = 5, figure_width_units = "inches", cmap = "gist_earth", hillshade = True, alpha_hillshade = HS_alpha, color_min = 0, color_max = mx, dpi = 500, output = "nothing", format_figure = "png", fontsize_ticks = 6 , fontsize_label = 8,hillshade_cmin = None, hillshade_cmax = None, fig = figt, colorbar = True)
            else:
                qp.plot_nice_topography(tdem ,figure_width = 5, figure_width_units = "inches", cmap = "gist_earth", hillshade = True, alpha_hillshade = HS_alpha, color_min = 0, color_max = mx, dpi = 500, output = "save", format_figure = "png", fontsize_ticks = 6 , fontsize_label = 8,hillshade_cmin = None, hillshade_cmax = None, colorbar = True)


        base_df = pd.read_csv(self.save_dir + self.prefix + "_model_info.csv")
        first_frame = base_df["Frame_num"].iloc[0] if(base_df["Frame_num"].iloc[0]!= -1) else base_df["Frame_num"].iloc[1]
        tdem = LSDDEM(path = self.save_dir, file_name = self.prefix+str(first_frame)+".bil",already_preprocessed = True)
        mx = np.percentile(tdem.cppdem.get_PP_raster(),90)
        fig,ax = qp.plot_nice_topography(tdem ,figure_width = 5, figure_width_units = "inches", cmap = "gist_earth", hillshade = True, alpha_hillshade = HS_alpha, color_min = 0, color_max = mx, dpi = 500, output = "return", format_figure = "png", fontsize_ticks = 6 , fontsize_label = 8,hillshade_cmin = None, hillshade_cmax = None, colorbar = True)
        moviewriter = FFMpegWriter(fps=fps)
        with moviewriter.saving(fig, self.save_dir+self.prefix+rename_suffix+".mp4", dpi=dpi):

            for j in base_df["Frame_num"]:
                if(j!=-1):
                    plotty(self.save_dir,"%s.bil"%(self.prefix+str(j)), True, fig,HS_alpha = HS_alpha)
                    moviewriter.grab_frame()
                    plt.clf()


# -*- coding: utf-8 -*-

"""Top-level package for LSDTopyTools."""

__author__ = """Boris Gailleton"""
__email__ = 'b.gailleton@sms.ed.ac.uk'
__version__ = '0.0.1'
import sys
# Dealing with the imports: making it compatible with python 3 and 2
if(sys.version[0] == 2):
	from lsdtt_concavity_tools import *
	from lsdtt_hypsometry import *
	from lsdtt_basic_operations import *

else:
	from .lsdtt_concavity_tools import *
	from .lsdtt_hypsometry import *
	from .lsdtt_basic_operations import *

